
# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Base class for the phenomenological noise models for inference."""

from typing import List, Tuple, Optional
from torch import Tensor
from torch.autograd import no_grad

from aihwkit.inference.converter.base import BaseConductanceConverter
from aihwkit.inference.converter.conductance import SinglePairConductanceConverter


class BaseNoiseModel:
    """Base class for phenomenological noise models for inference."""

    def __init__(
            self,
            g_converter: Optional[BaseConductanceConverter] = None
    ):
        self.g_converter = g_converter or SinglePairConductanceConverter()

    @no_grad()
    def apply_noise(self, weights: Tensor, t_inference: float) -> Tensor:
        """Apply the expected noise.

        Applies the noise to a non-perturbed conductance matrix ``weights``
        at time of inference ``t_inference`` (in seconds) where 0 sec
        refers to the time when weight programming has finished.

        Note:
            The drift coefficients and intermediate noises etc. are
            sampled for each application of this function anew from the
            distributions, thus it samples the expected noise and drift
            behavior at time ``t_inference`` but not a continual
            trajectory of a given device instance over time (having
            e.g. constant drift coefficients).
        """
        target_conductances, params = self.g_converter.convert_to_conductances(weights)

        noisy_conductances = []
        for g_target in target_conductances:
            g_prog = self.apply_programming_noise_to_conductance(g_target)
            if t_inference > 0:
                nu_drift = self.generate_drift_coefficients(g_target)
                noisy_conductances.append(self.apply_drift_noise_to_conductance(
                    g_prog, nu_drift, t_inference))

        noisy_weights = self.g_converter.convert_back_to_weights(noisy_conductances, params)

        return noisy_weights

    @no_grad()
    def apply_programming_noise(self, weights: Tensor) -> Tuple[Tensor, List[Tensor]]:
        """Apply the expected programming noise to weights.

        Uses the :meth:`~apply_programming_noise_to_conductances` on
        each of the conductance slices.

        Args:
            weights: weights tensor

        Returns:
            weight tensor with programming noise applied, and tuple of
            all drift coefficients (per conductances slice) that are
            determined during programming.
        """
        target_conductances, params = self.g_converter.convert_to_conductances(weights)

        noisy_conductances = []
        nu_drift_list = []
        for g_target in target_conductances:

            noisy_conductances.append(self.apply_programming_noise_to_conductance(g_target))
            nu_drift_list.append(self.generate_drift_coefficients(g_target))
        noisy_weights = self.g_converter.convert_back_to_weights(noisy_conductances, params)

        return noisy_weights, nu_drift_list

    @no_grad()
    def apply_drift_noise(
            self,
            weights: Tensor,
            nu_drift_list: List[Tensor],
            t_inference: float
    ) -> Tensor:
        """Apply the expected drift noise to weights.

        Uses the :meth:`~apply_drift_noise_to_conductances` on
        each of the conductance slices.

        Args:
            weights: weights tensor (usually with programming noise already applied)
            nu_drift_list: list of drift nu for each conductance slice
            t_inference: assumed time of inference (in sec)

        Returns:
            weight tensor with drift noise applied
        """
        target_conductances, params = self.g_converter.convert_to_conductances(weights)

        noisy_conductances = []
        for g_target, nu_drift in zip(target_conductances, nu_drift_list):
            noisy_conductances.append(
                self.apply_drift_noise_to_conductance(g_target, nu_drift, t_inference))

        noisy_weights = self.g_converter.convert_back_to_weights(noisy_conductances, params)

        return noisy_weights

    @no_grad()
    def generate_drift_coefficients(self, g_target: Tensor) -> Tensor:
        """Generate drift coefficients ``nu`` based on the target conductances."""
        raise NotImplementedError

    @no_grad()
    def apply_programming_noise_to_conductance(self, g_target: Tensor) -> Tensor:
        r"""Apply programming noise to a target conductance ``Tensor``.

        Args:
            g_target: Target conductances

        Returns:
            Tensor of sampled drift coefficients :math:`\nu`, one for each
            target conductance value.
        """
        raise NotImplementedError

    @no_grad()
    def apply_drift_noise_to_conductance(
            self,
            g_prog: Tensor,
            nu_drift: Tensor,
            t_inference: float
    ) -> Tensor:
        r"""Apply the noise and drift up to the assumed inference time point.

        Args:
            g_prog: Tensor of conductance values after programming (in :math:`\muS`)
            nu_drift: drift nu
            t_inference: assumed time of inference (in sec)

        Returns:
            conductance Tensor with applied noise and drift
        """
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Helper for retrieving Metrics of an Experiment."""

from datetime import datetime
import json
from typing import Dict


class InferenceLocalMetric:
    """Metric used by the InferenceWorker Runner."""

    def __init__(self, stdout: bool = False) -> None:
        self.current_repeat: Dict = {}
        self.time_init = datetime.utcnow()
        self.stdout = stdout

    def receive_repeat_start(self, repeat: int) -> None:
        """Hook for `INFERENCE_REPEAT_START`."""
        self.current_repeat = {
                "number": repeat,
                "inference_results": []
        }

    def receive_repeat_end(self, t_inference_array: list, avg_acc_arr: list,
                           std_acc_arr: list, avg_err_arr: list, avg_loss_arr: list,
                           inference_repeats: int) -> Dict:
        """Hook for `INFERENCE_REPEAT_END`."""

        inf_results = []
        n_inference = len(t_inference_array)

        # The input are the arrays of avg accuracy, avg error and avg loss.
        # Create the dict entry for the items in the arrays.
        for i in range(n_inference):
            new_dict = {
                    "t_inference": t_inference_array[i],
                    "avg_accuracy": avg_acc_arr[i],
                    "std_accuracy": std_acc_arr[i],
                    "avg_error": avg_err_arr[i],
                    "avg_loss": avg_loss_arr[i]}
            inf_results.append(new_dict)

        repeat = self.current_repeat["number"] + 1
        time_elapsed = (datetime.utcnow() - self.time_init).total_seconds()
        is_partial = bool(repeat < inference_repeats)
        partial = {
            "inference_runs": {
                    "inference_repeat": repeat,
                    "is_partial": is_partial,
                    "time_elapsed": time_elapsed,
                    "inference_results":  inf_results
            }
        }

        if self.stdout:
            print("{}".format(json.dumps(partial)))

        # Return the partial.
        return partial



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=too-many-instance-attributes

"""Utility enumerators for resistive processing units."""

from enum import Enum


class BoundManagementType(Enum):
    """Bound management type.

    In the case ``Iterative`` the MAC is iteratively recomputed with
    inputs iteratively halved, when the output bound was hit.

    Caution:
        Bound management is **only** available for the forward pass. It
        will be ignored when used for the backward pass.
    """

    NONE = 'None'
    """No bound management."""

    ITERATIVE = 'Iterative'
    r"""Iteratively recomputes input scale set to :math:`\alpha\leftarrow\alpha/2`.

    It iteratively recomputes the bounds up to limit of passes (given by
    ``max_bm_factor`` or ``max_bm_res``).
    """

    ITERATIVE_WORST_CASE = 'IterativeWorstCase'
    """Worst case bound management.

    Uses ``AbsMax`` noise management for the first pass and only when output
    bound is hit, the ``AbsMaxNPSum`` for the second. Thus, at most 2 passes
    are computed.
    """

    SHIFT = 'Shift'
    """Shift bound management.

    Shifts the output by adding the difference ``output_bound - max_output`` to
    the analog output value. This is only useful to increase the dynamic range
    before the softmax, where the max can be safely.

    Note:
        Shifting needs hardware implementations.
    """


class NoiseManagementType(Enum):
    r"""Noise management type.

    Noise management determines a factor :math:`\alpha` how the input is reduced:

    .. math:: \mathbf{y} = \alpha\;F_\text{analog-mac}\left(\mathbf{x}/\alpha\right)
    """

    NONE = 'None'
    """No noise management."""

    ABS_MAX = 'AbsMax'
    r"""Use :math:`\alpha\equiv\max{|\mathbf{x}|}`."""

    ABS_MAX_NP_SUM = 'AbsMaxNPSum'
    """Assume weight value is constant and given by ``nm_assumed_wmax``.

    Takes a worst case scenario of the weight matrix to calculate the input
    scale to ensure that output is not clipping. Assumed weight value is
    constant and given by ``nm_assumed_wmax``.
    """

    MAX = 'Max'
    r"""Use :math:`\alpha\equiv\max{\mathbf{x}}`."""

    CONSTANT = 'Constant'
    r"""A constant value (given by parameter ``nm_thres``)."""

    AVERAGE_ABS_MAX = 'AverageAbsMax'
    """Moment-based scale input scale estimation.

    Computes the average abs max over the mini-batch and applies ``nm_decay``
    to update the value with the history.

    Note:
        ``nm_decay`` is ``1-momentum`` and always given in mini-batches.
        However, the CUDA implementation does not discount values within
        mini-batches, whereas the CPU implementation does.
    """


class WeightNoiseType(Enum):
    r"""Output weight noise type.

    The weight noise is applied for each MAC computation, while not
    touching the actual weight matrix but referring it to the output.

    .. math:: y_i = \sum_j w_{ij}+\xi_{ij}
    """

    NONE = 'None'
    """No weight noise."""

    ADDITIVE_CONSTANT = 'AdditiveConstant'
    r"""The :math:`\xi\sim{\cal N}(0,\sigma)` thus all are Gaussian distributed.

    :math:`\sigma` is determined by ``w_noise``.
    """

    PCM_READ = 'PCMRead'
    """Output-referred PCM-like read noise.

    Output-referred PCM-like read noise that scales with the amount of current
    generated for each output line and thus scales with both conductance values
    and input strength.

    The same general for is taken as for PCM-like statistical model of the 1/f
    noise during inference, see
    :class:`aihwkit.inference.noise.pcm.PCMLikeNoiseModel`.
    """


class PulseType(Enum):
    """Pulse type."""

    NONE = 'None'
    """Floating point update instead of pulses."""

    STOCHASTIC_COMPRESSED = 'StochasticCompressed'
    """Generates actual stochastic bit lines.

    Plus and minus pulses are taken in the same pass.
    """

    STOCHASTIC = 'Stochastic'
    """Two passes for plus and minus (only CPU)."""

    NONE_WITH_DEVICE = 'NoneWithDevice'
    """Floating point like ``None``, but with analog devices (e.g. weight
    clipping)."""

    MEAN_COUNT = 'MeanCount'
    """Coincidence based in prob (:math:`p_a p_b`)."""

    DETERMINISTIC_IMPLICIT = 'DeterministicImplicit'
    r"""Coincidences are computed in deterministic manner.

    Coincidences are calculated by :math:`b_l x_q d_q` where ``BL`` is the
    desired bit length (possibly subject to dynamic adjustments using
    ``update_bl_management``) and :math:`x_q` and :math:`d_q` are the quantized
    input and error values, respectively, normalized to the range
    :math:`0,\ldots,1`. It can be shown that explicit bit lines exist that
    generate these coincidences.
    """


class WeightModifierType(Enum):
    """Weight modifier type."""

    COPY = 'Copy'
    """Just copy, however, could also drop."""

    DISCRETIZE = 'Discretize'
    """Quantize the weights."""

    MULT_NORMAL = 'MultNormal'
    """Multiplicative Gaussian noise."""

    ADD_NORMAL = 'AddNormal'
    """Additive Gaussian noise."""

    DISCRETIZE_ADD_NORMAL = 'DiscretizeAddNormal'
    """First discretize and then additive Gaussian noise."""

    DOREFA = 'DoReFa'
    """DoReFa discretization."""

    POLY = 'Poly'
    r"""Nth order Polynomial noise model (in terms of the weight value).

    In detail, for the duration of a mini-batch, each weight will be
    added a Gaussian random number with the standard deviation of
    :math:`\sigma_\text{wnoise} (c_0 + c_1 w_{ij}/\omega +
    c_N w_{ij}^N/\omega^N)`, where :math:`omega` is either the actual
    max weight (if ``rel_to_actual_wmax`` is set) or the value
    ``assumed_wmax``.
    """


class WeightClipType(Enum):
    """Weight clipper type."""

    NONE = 'None'
    """None."""

    FIXED_VALUE = 'FixedValue'
    """Clip to fixed value give, symmetrical around zero."""

    LAYER_GAUSSIAN = 'LayerGaussian'
    """Calculates the second moment of the whole weight matrix and clips
    at ``sigma`` times the result symmetrically around zero."""

    AVERAGE_CHANNEL_MAX = 'AverageChannelMax'
    """Calculates the abs max of each output channel (row of the weight
    matrix) and takes the average as clipping value for all."""


class WeightRemapType(Enum):
    """Weight clipper type."""

    NONE = 'None'
    """None."""

    LAYERWISE_SYMMETRIC = 'LayerwiseSymmetric'
    """Remap according to the absolute max of the full weight matrix."""

    CHANNELWISE_SYMMETRIC = 'ChannelwiseSymmetric'
    """Remap each column (output channel) in respect to the absolute max."""


class VectorUnitCellUpdatePolicy(Enum):
    """Vector unit cell update policy."""

    ALL = 'All'
    """All devices updated simultaneously."""

    SINGLE_FIXED = 'SingleFixed'
    """Device index is not changed. Can be set initially and/or updated on
    the fly."""

    SINGLE_SEQUENTIAL = 'SingleSequential'
    """Each device one at a time in sequence."""

    SINGLE_RANDOM = 'SingleRandom'
    """A single device is selected by random choice each mini-batch."""


class AnalogMVType(Enum):
    """Type of the analog matrix-vector product."""

    IDEAL = 'Ideal'
    """FP mat-vec without any non-idealities. Same as setting
    ``is_perfect=True``."""

    ONE_PASS = 'OnePass'
    """One pass through the crossbar array for positive and negative inputs."""

    POS_NEG_SEPARATE = 'PosNegSeparate'
    """Two passes through the crossbar array for positive and negative
    inputs separately. The output of the two passes are added in
    analog and then passed once through ADC stage (which also applies
    the output noise, range clipping, output non-linearity etc.).
    """

    POS_NEG_SEPARATE_DIGITAL_SUM = 'PosNegSeparateDigitalSum'
    """Two passes through the crossbar array for positive and negative
    inputs separately. The ADC output stage is applied to each pass
    separately and the results are summed in full precision (i.e. in
    digital).
    """



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Analog Modules that contain children Modules."""

from typing import (
    Callable, Optional, Union, Any, NamedTuple, Tuple, TYPE_CHECKING, Generator
)
from collections import OrderedDict

from torch import device as torch_device
from torch.nn import Sequential

from aihwkit.exceptions import ModuleError, TileError
from aihwkit.nn.modules.base import AnalogModuleBase

if TYPE_CHECKING:
    from torch import Tensor  # pylint: disable=ungrouped-imports


class AnalogSequential(Sequential):
    """An analog-aware sequential container.

    Specialization of torch ``nn.Sequential`` with extra functionality for
    handling analog layers:

    * correct handling of ``.cuda()`` for children modules.
    * apply analog-specific functions to all its children (drift and program
      weights).

    Note:
        This class is recommended to be used in place of ``nn.Sequential`` in
        order to correctly propagate the actions to all the children analog
        layers. If using regular containers, please be aware that operations
        need to be applied manually to the children analog layers when needed.
    """
    # pylint: disable=abstract-method

    def _apply_to_analog(self, fn: Callable) -> 'AnalogSequential':
        """Apply a function to all the analog layers in this module.

        Args:
            fn: function to be applied.

        Returns:
            This module after the function has been applied.
        """
        for module in self.analog_modules():
            fn(module)

        return self

    def apply_to_analog_modules(self, fn: Callable) -> 'AnalogSequential':
        """Apply a function to all the analog modules.

        Args:
            fn: function to be applied.

        Returns:
            This module after the function has been applied.
        """
        for module in self.analog_modules():
            fn(module)

        return self

    def apply_to_analog_tiles(self, fn: Callable) -> 'AnalogSequential':
        """Apply a function to all the analog tiles of all layers in this module.

        Example::

            model.apply_to_analog_tiles(lambda tile: tile.reset())

        This would reset each analog tile in the whole DNN looping
        through all layers and all tiles that might exist in a
        particular layer.

        Args:
            fn: function to be applied.

        Returns:
            This module after the function has been applied.

        """
        for module in self.analog_modules():
            for analog_tiles in module.analog_tiles():
                fn(analog_tiles)
        return self

    def analog_modules(self) -> Generator[AnalogModuleBase, None, None]:
        """Generator over analog modules only"""
        for module in self.modules():
            if isinstance(module, AnalogModuleBase):
                yield module

    def named_analog_modules(self) -> Generator[Tuple[str, AnalogModuleBase], None, None]:
        """Generator over analog modules only"""
        for name, module in self.named_modules():
            if isinstance(module, AnalogModuleBase):
                yield name, module

    def cpu(
            self
    ) -> 'AnalogSequential':
        super().cpu()

        self._apply_to_analog(lambda m: m.cpu())

        return self

    def cuda(
            self,
            device: Optional[Union[torch_device, str, int]] = None
    ) -> 'AnalogSequential':
        super().cuda(device)

        self._apply_to_analog(lambda m: m.cuda(device))

        return self

    def to(
            self,
            device: Optional[Union[torch_device, str, int]] = None
    ) -> 'AnalogSequential':
        """Move and/or cast the parameters, buffers and analog tiles.

        Note:
            Please be aware that moving analog layers from GPU to CPU is
            currently not supported.

        Args:
            device: the desired device of the parameters, buffers and analog
                tiles in this module.

        Returns:
            This module in the specified device.
        """
        # pylint: disable=arguments-differ
        device = torch_device(device)

        super().to(device)

        if device.type == 'cuda':
            self._apply_to_analog(lambda m: m.cuda(device))
        elif device.type == 'cpu':
            self._apply_to_analog(lambda m: m.cpu())

        return self

    def get_analog_tile_device(self) -> Union[torch_device, str, int]:
        """ Return the devices used by the analog tiles.

        Returns:
            device: torch device

        Raises:
            TileError: in case the model is on non-unique devices
        """

        devices = []
        dev_name = []

        self._apply_to_analog(
            lambda mod: devices.extend(mod.get_analog_tile_devices()))

        for name in devices:
            dev_name.append(str(name))

        if len(set(dev_name)) > 1:
            raise TileError("Torch device is not unique")

        device = devices[0]

        return device

    def load_state_dict(self,  # pylint: disable=arguments-differ
                        state_dict: 'OrderedDict[str, Tensor]',
                        strict: bool = True,
                        load_rpu_config: bool = True) -> NamedTuple:
        """Specializes torch's ``load_state_dict`` to add a flag whether to
        load the RPU config from the saved state.

        Args:
            state_dict: see torch's ``load_state_dict``
            strict: see torch's ``load_state_dict``
            load_rpu_config: Whether to load the saved RPU
                config or use the current RPU config of the model.

                Caution:

                    If ``load_rpu_config=False`` the RPU config can
                    be changed from the stored model. However, the user has to
                    make sure that the changed RPU config makes sense.

                    For instance, changing the device type might
                    change the expected fields in the hidden
                    parameters and result in an error.
        Returns:
            see torch's ``load_state_dict``

        Raises: ModuleError: in case the rpu_config class mismatches
            for ``load_rpu_config=False``.
        """
        # pylint: disable=protected-access
        self._apply_to_analog(lambda m: m._set_load_rpu_config_state(load_rpu_config))
        return super().load_state_dict(state_dict, strict)

    def prepare_for_ddp(self) -> None:
        """Adds ignores to avoid broadcasting the analog tile states in case of
        distributed training.

        Note:
            Call this function before the mode is converted with DDP.

        Important:
            Only InferenceTile supports DDP.

        Raises:

            ModuleError: In case analog tiles are used that do not
                support data-parallel model, ie. all analog training
                tiles.
        """
        # pylint: disable=attribute-defined-outside-init
        exclude_list = []
        for module in self.modules():
            if isinstance(module, AnalogModuleBase):
                for analog_tile in module.analog_tiles():
                    if analog_tile.shared_weights is None:
                        raise ModuleError("DDP is only supported with shared weights"
                                          "(e.g. InferenceTile)")
                exclude_list += [module.ANALOG_CTX_PREFIX, module.ANALOG_STATE_PREFIX]
        exclude_list = list(set(exclude_list))
        params = self.state_dict().keys()
        exclude_params = []
        for param in params:
            for word in exclude_list:
                if word in param and word not in exclude_params:
                    exclude_params.append(param)
                    break
        self._ddp_params_and_buffers_to_ignore = exclude_params

    def remap_analog_weights(self, weight_scaling_omega: Optional[float] = 1.0) -> None:
        """Remap the analog weights and set the digital out scales.

        Caution:

            This should typically *not* be called for analog training
            unless realistic_read_write is set. In this case, it would
            perform a full re-write of the weights. However,
            typically, this this method is intended to correct the
            mapping for hardware-aware trained models before doing the
            inference with programmed weights.

        Args:

            weight_scaling_omega: The optional value to remap the
                weight max to. If None it will take the value set
                initially in the ``RPUConfig.mapping``. Defaults to 1.0.

        """

        self._apply_to_analog(lambda m: m.remap_weights(
            weight_scaling_omega=weight_scaling_omega))

    def drift_analog_weights(self, t_inference: float = 0.0) -> None:
        """(Program) and drift all analog inference layers of a given model.

        Args:
            t_inference: assumed time of inference (in sec)

        Raises:
            ModuleError: if the layer is not in evaluation mode.
        """
        if self.training:
            raise ModuleError('drift_analog_weights can only be applied in '
                              'evaluation mode')

        self._apply_to_analog(lambda m: m.drift_analog_weights(t_inference))

    def program_analog_weights(self) -> None:
        """Program all analog inference layers of a given model.

        Raises:
            ModuleError: if the layer is not in evaluation mode.
        """
        if self.training:
            raise ModuleError('program_analog_weights can only be applied in '
                              'evaluation mode')

        self._apply_to_analog(lambda m: m.program_analog_weights())

    @classmethod
    def from_digital(cls, module: Sequential,  # pylint: disable=unused-argument
                     *args: Any,
                     **kwargs: Any) -> 'AnalogSequential':
        """Construct AnalogSequential in-place from Sequential."""
        return cls(OrderedDict(mod for mod in module.named_children()))



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Visualization utilities (web)."""

import argparse
from typing import Optional
from pathlib import Path

from cycler import cycler
import matplotlib.pyplot as plt

from matplotlib.axes import Axes
from matplotlib.figure import Figure

from aihwkit.simulator.presets.devices import (
    ReRamSBPresetDevice, ReRamESPresetDevice, CapacitorPresetDevice,
    EcRamPresetDevice, IdealizedPresetDevice
)
from aihwkit.simulator.presets.compounds import PCMPresetUnitCell
from aihwkit.simulator.configs.devices import PulsedDevice
from aihwkit.utils.visualization import plot_device_compact

# Colors used by the frontend.
WEB_COLORS = ['#8A3FFC', '#33B1FF', '#007D79', '#FF7EB6', '#FA4D56', '#FFF1F1',
              '#6FDC8C', '#4589FF', '#D12771', '#D2A106', '#08BDBA', '#BAE6FF',
              '#BA4E00', '#D4BBFF']
# Devices for which plots should be generated.
DEVICES = {
    ReRamESPresetDevice: 1000,
    ReRamSBPresetDevice: 1000,
    CapacitorPresetDevice: 400,
    EcRamPresetDevice: 1000,
    IdealizedPresetDevice: 10000,
    PCMPresetUnitCell: 80
}


def set_dark_style(axes: Axes) -> None:
    """Sets a nice color cycle for a given axes."""

    axes.set_prop_cycle(cycler(color=WEB_COLORS))

    axes.set_facecolor('#262626')


def plot_device_compact_web(
        device: PulsedDevice,
        w_noise: float = 0.0,
        n_steps: Optional[int] = None,
        n_traces: int = 3
) -> Figure:
    """Plots a compact step response figure for a given device (preset).

    Note:
        It will use an amount of read weight noise ``w_noise`` for
        reading the weights.

    Params:
        device: PulsedDevice parameters
        w_noise: Weight noise standard deviation during read
        n_steps: Number of steps for up/down cycle
        n_traces: Number of traces to plot (for device-to-device variation)
        show: if `True`, displays the figure.

    Returns:
        the compact step response figure.
    """
    plt.style.use('dark_background')

    figure = plot_device_compact(device, w_noise, n_steps, n_traces)

    # Tune for web.
    for axes in figure.axes:
        for i, line in enumerate(axes.lines):
            line.set_color(WEB_COLORS[i])
        # set_dark_style(axes)

    return figure


def save_plots_for_web(
        path: Path = Path('/tmp'),
        file_format: str = 'svg'
) -> None:
    """Create the plots for the web.

    Args:
        path: the path where the images will be stored.
        file_format: the image format.
    """
    def camel_to_snake(source: str) -> str:
        """Convert a CamelCase string into snake-case."""
        return ''.join(['_' + char.lower() if char.isupper() else char
                        for char in source]).lstrip('_')

    for device, n_steps in DEVICES.items():
        # Images for the detailed modal.
        file_name = '{}.{}'.format(camel_to_snake(device.__name__), file_format)
        file_path = path.absolute() / file_name

        figure = plot_device_compact_web(device(), n_steps=n_steps)
        figure.savefig(file_path, format=file_format, transparent=True,
                       bbox_inches='tight')

        # Images for the mini leftbar.
        file_name = '{}-mini.{}'.format(camel_to_snake(device.__name__), file_format)
        file_path = path.absolute() / file_name

        figure = plot_device_compact_web(device(), n_traces=1, n_steps=n_steps)
        for axes in figure.axes:
            # Disable texts.
            axes.set_title('')
            axes.set_xlabel('')
            axes.set_ylabel('')
            # Disable tick labels.
            axes.xaxis.set_ticklabels([])
            axes.yaxis.set_ticklabels([])
            # Disable axis entirely.
            axes.get_xaxis().set_visible(False)
            axes.get_yaxis().set_visible(False)
            # Increase axis width.
            for axis in ['top', 'bottom', 'left', 'right']:
                axes.spines[axis].set_linewidth(3)
            for line in axes.lines:
                line.set_linewidth(4)
        figure.savefig(file_path, format=file_format, transparent=True,
                       bbox_inches='tight')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Generate plots for frontend.')
    parser.add_argument('destination', type=str,
                        help='folder where the plots will be stored')
    args = parser.parse_args()

    destination_path = Path(args.destination)
    destination_path.mkdir(exist_ok=True)
    save_plots_for_web(destination_path)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022, 2023 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Converters for `BasicInferencing` Experiment."""

from typing import Any, Dict, List
# from collections import OrderedDict

from torch.nn import Module, NLLLoss

from aihwkit.simulator.configs import InferenceRPUConfig
from aihwkit.cloud.converter.exceptions import ConversionError
from aihwkit.cloud.converter.v1.i_mappings import (
    InverseMappings, Mappings)

# pylint: disable=no-name-in-module,import-error
from aihwkit.experiments.experiments.inferencing import (  # type: ignore[import]
    BasicInferencing
)
from aihwkit.cloud.converter.definitions.i_input_file_pb2 import (   # type: ignore[attr-defined]
    InferenceInput,
    Dataset,
    Inferencing,
    NoiseModelProto,
    PCMProto, GenericProto,
    AnalogProto
)

from aihwkit.cloud.converter.definitions.i_common_pb2 import (   # type: ignore[attr-defined]
    LayerOrActivationFunction,
    Network,
    LayerProto,
    ActivationFunctionProto,
    Version
)
from aihwkit.cloud.converter.definitions.i_output_file_pb2 import (  # type: ignore[attr-defined]
    InferenceRunsProto, InferenceResultsProto, InferencingOutput
)

from aihwkit.nn import AnalogSequential

from aihwkit.cloud.converter.v1.analog_info import AnalogInfo
from aihwkit.cloud.converter.v1.noise_model_info import NoiseModelInfo
from aihwkit.cloud.converter.v1.rpu_config_info import RPUconfigInfo


class BasicInferencingConverter:
    """Converter for `BasicInferencing` Experiment."""

    def to_proto(self, experiment: BasicInferencing,
                 analog_info: Dict,
                 noise_model_info: Dict) -> Any:
        """Convert an `Experiment` to its protobuf representation."""

        version = self._version_to_proto()
        dataset = self._dataset_to_proto(experiment.dataset,
                                         experiment.batch_size)
        network = self._model_to_proto(experiment.model,
                                       experiment.weight_template_id)
        inferencing = self._inferencing_to_proto(experiment.inference_repeats,
                                                 experiment.inference_time,
                                                 analog_info,
                                                 noise_model_info,
                                                 )
        return InferenceInput(
            version=version,
            dataset=dataset,
            network=network,
            inferencing=inferencing
        )

    def from_proto(self, protobuf: Any) -> BasicInferencing:
        """Convert a protobuf representation to an `Experiment`."""

        dataset = InverseMappings.datasets[protobuf.dataset.dataset_id]
        layers = protobuf.network.layers
        # build RPUconfig_info to be used when it is instantiated dynamically
        alog_info = AnalogInfo(protobuf.inferencing.analog_info)
        nm_info = NoiseModelInfo(protobuf.inferencing.noise_model_info)
        rc_info = RPUconfigInfo(nm_info, alog_info, layers)

        model = self._model_from_proto(protobuf.network, rc_info)

        batch_size = protobuf.dataset.batch_size
        inference_info = self._inference_info_from_proto(protobuf)
        loss_function = NLLLoss
        weight_template_id = inference_info['weight_template_id']
        inference_repeats = inference_info['inference_repeats']
        inference_time = inference_info['inference_time']

        return BasicInferencing(
            dataset=dataset,
            model=model,
            batch_size=batch_size,
            loss_function=loss_function,
            weight_template_id=weight_template_id,
            inference_repeats=inference_repeats,
            inference_time=inference_time
        )

    # Methods for converting to proto.

    @staticmethod
    def _version_to_proto() -> Any:
        return Version(
            schema=1,
            opset=1
        )

    @staticmethod
    def _dataset_to_proto(dataset: type, batch_size: int) -> Any:
        if dataset not in Mappings.datasets:
            raise ConversionError(f'Unsupported dataset: {dataset}')

        return Dataset(
            dataset_id=Mappings.datasets[dataset],
            batch_size=batch_size
        )

    @staticmethod
    def _model_to_proto(model: Module, weight_template_id: str) -> Any:

        if not isinstance(model, AnalogSequential):
            raise ConversionError('Unsupported model: only AnalogSequential is supported')

        children_types = {type(layer) for layer in model.children()}
        valid_types = set(Mappings.layers.keys()) | set(Mappings.activation_functions.keys())
        if children_types - valid_types:
            raise ConversionError('Unsupported layers: '
                                  f'{children_types - valid_types}')

        # Create a new input_file pb Network object with weight_template_id
        network = Network(weight_template_id=weight_template_id)

        for child in model.children():
            child_type = type(child)
            if child_type in Mappings.layers:
                item = LayerOrActivationFunction(
                    layer=Mappings.layers[child_type].to_proto(child, LayerProto)
                )
            else:
                item = LayerOrActivationFunction(
                    activation_function=Mappings.activation_functions[child_type].to_proto(
                        child, ActivationFunctionProto)
                )
            # pylint: disable=no-member),undefined-variable
            network.layers.extend([item])

        return network

    @staticmethod
    def _noise_model_to_proto(
            noise_model_info: Dict) -> NoiseModelProto:  # type: ignore[valid-type]
        """Creates a protobuf NoiseModelProto object from input dictionaries"""

        model = None
        device_id = noise_model_info['device_id']
        if device_id == 'pcm':
            model = NoiseModelProto(
                pcm=PCMProto(
                    device_id=device_id,
                    programming_noise_scale=noise_model_info['programming_noise_scale'],
                    read_noise_scale=noise_model_info['read_noise_scale'],
                    drift_scale=noise_model_info['drift_scale'],
                    drift_compensation=noise_model_info['drift_compensation'],
                    poly_first_order_coef=noise_model_info['poly_first_order_coef'],
                    poly_second_order_coef=noise_model_info['poly_second_order_coef'],
                    poly_constant_coef=noise_model_info['poly_constant_coef']
                    )
                )
        else:
            model = NoiseModelProto(
                generic=GenericProto(
                    device_id=device_id,
                    programming_noise_scale=noise_model_info['programming_noise_scale'],
                    read_noise_scale=noise_model_info['read_noise_scale'],
                    drift_scale=noise_model_info['drift_scale'],
                    drift_compensation=noise_model_info['drift_compensation'],
                    poly_first_order_coef=noise_model_info['poly_first_order_coef'],
                    poly_second_order_coef=noise_model_info['poly_second_order_coef'],
                    poly_constant_coef=noise_model_info['poly_constant_coef'],
                    drift_mean=noise_model_info['drift_mean'],
                    drift_std=noise_model_info['drift_std'],
                    )
            )
        return model

    @staticmethod
    def rpu_config_info_from_info(analog_info: Dict,
                                  noise_model_info: Dict
                                  ) -> RPUconfigInfo:
        """ Creates RPUconfigInfo """

        # TODO: Remove the "Info" objects: NoiseModelInfo, AnalogInfo. Seems unecessary
        nm_info = NoiseModelInfo(BasicInferencingConverter._noise_model_to_proto(
            noise_model_info))  # type: ignore[name-defined]
        a_info = AnalogInfo(AnalogProto(**analog_info))
        return RPUconfigInfo(nm_info, a_info, None)

    @staticmethod
    def rpu_config_from_info(analog_info: Dict,
                             noise_model_info: Dict,
                             func_id: str = 'id-not-provided'
                             ) -> InferenceRPUConfig:
        """ Creates RPUConfig """
        nm_info = NoiseModelInfo(BasicInferencingConverter._noise_model_to_proto(
            noise_model_info))  # type: ignore[name-defined]
        a_info = AnalogInfo(AnalogProto(**analog_info))
        return RPUconfigInfo(nm_info,
                             a_info, None).create_inference_rpu_config(func_id)

    @staticmethod
    def _inferencing_to_proto(
            inference_repeats: int,
            inference_time: float,
            analog_info: Dict,
            noise_model_info: Dict) -> Inferencing:  # type: ignore[valid-type]
        """Creates protobuf Inferencing object"""

        # Not sure why mypy and pylint cannot find following method, python3 can.
        # pylint: disable=undefined-variable
        nm_info = BasicInferencingConverter._noise_model_to_proto(
            noise_model_info)  # type: ignore[name-defined]
        a_info = AnalogProto(**analog_info)

        return Inferencing(inference_repeats=inference_repeats,
                           inference_time=inference_time,
                           analog_info=a_info,
                           noise_model_info=nm_info
                           )

    # Methods for converting from proto.

    @staticmethod
    def _inference_info_from_proto(
            inference_pb: InferenceInput) -> Dict:  # type: ignore[valid-type]
        """Converts inference_info from protobuf to a dictionary"""

        inferencing = inference_pb.inferencing  # type: ignore[attr-defined]
        network = inference_pb.network  # type: ignore[attr-defined]
        return {
                    'inference_repeats': inferencing.inference_repeats,
                    'inference_time': inferencing.inference_time,
                    'weight_template_id': network.weight_template_id
                }

    @staticmethod
    def _model_from_proto(network: Network,  # type: ignore[valid-type]
                          rc_info: RPUconfigInfo) -> Module:
        layers = []
        for layer_proto in network.layers:  # type: ignore[attr-defined]
            if layer_proto.WhichOneof('item') == 'layer':
                layer_cls = InverseMappings.layers[layer_proto.layer.id]
                layer = Mappings.layers[layer_cls].from_proto(
                    layer_proto.layer, layer_cls, {'rpu_config': rc_info})
            else:
                layer_cls = InverseMappings.activation_functions[layer_proto.activation_function.id]
                layer = Mappings.activation_functions[layer_cls].from_proto(
                    layer_proto.activation_function, layer_cls)

            layers.append(layer)

        return AnalogSequential(*layers)

    @staticmethod
    def _analog_info_from_proto(
            analog_info: AnalogProto) -> Dict:  # type: ignore[valid-type]
        """Converts from protobuf analog_info to a dictionary"""

        return {
                    'output_noise_strength': (
                        analog_info.output_noise_strength  # type: ignore[attr-defined]
                    ),
                    'adc': analog_info.adc,  # type: ignore[attr-defined]
                    'dac': analog_info.dac  # type: ignore[attr-defined]
                }

    @staticmethod
    def _noise_model_from_proto(
            noise_model_info: NoiseModelProto) -> Dict:  # type: ignore[valid-type]
        """Converts from protobuf noise_model to a dictionary"""

        extra = {}
        typ = noise_model_info.WhichOneof('item')  # type: ignore[attr-defined]

        if typ == 'pcm':
            # pcm does not have 2 extra fields
            info = noise_model_info.pcm  # type: ignore[attr-defined]

        elif typ == 'generic':
            info = noise_model_info.generic  # type: ignore[attr-defined]
            # There are 2 extra fields in generic
            extra = {
                'drift_mean': info.drift_mean,
                'drift_std': info.drift_std
            }
        else:
            raise TypeError

        base = {
            'device_id': typ,
            'programming_noise_scale': info.programming_noise_scale,
            'read_noise_scale': info.read_noise_scale,
            'drift_scale': info.drift_scale,
            'drift_compensation': info.drift_compensation,
            'poly_first_order_coef': info.poly_first_order_coef,
            'poly_second_order_coef': info.poly_second_order_coef,
            'poly_constant_coef': info.poly_constant_coef,
        }

        # add the extra fields if any in return value
        return {**base, **extra}


class BasicInferencingResultConverter:
    """Converter for `BasicInferencing` results."""

    def to_proto(self, results: Dict) -> Any:
        """Convert a result to its InferenceOutput object in i_output_file protobuf"""

        version = self._version_to_proto()
        inference_runs = self._runs_to_proto(results['inference_runs'])

        return InferencingOutput(
            version=version,
            inference_runs=inference_runs
        )

    @staticmethod
    def to_json_from_pb(inference_input: Any) -> Dict:
        """Convert a result to its json representation (inverse of to_proto())"""

        # Create an InferenceRunsProto object
        i_runs = inference_input.inference_runs  # type: ignore

        results = []        # this is a list

        # loop through all the results and append directly to InferenceRunsProto field
        for result in i_runs.inference_results:
            results.append(
                {
                    't_inference': result.t_inference,
                    'avg_accuracy': result.avg_accuracy,
                    'std_accuracy': result.std_accuracy,
                    'avg_error': result.avg_error,
                    'avg_loss': result.avg_loss
                }
            )

        # need to add 'inference_runs' dictionary key and value because to_proto() input
        #   contained a leading index.
        inference_runs = {
            'inference_runs': {
                'inference_repeat': i_runs.inference_repeat,
                'is_partial': i_runs.is_partial,
                'time_elapsed': i_runs.time_elapsed,
                'inference_results': results
            }
        }
        return inference_runs

    @staticmethod
    def result_from_proto(inference_input: Any) -> List[Dict]:
        """Convert a result to its json representation (inverse of to_proto())"""

        # Create an InferenceRunsProto object
        i_runs = inference_input.inference_runs  # type: ignore

        results = []        # this is a list

        # loop through all the results and append directly to InferenceRunsProto field
        for result in i_runs.inference_results:
            results.append(
                {
                    't_inference': result.t_inference,
                    'avg_accuracy': result.avg_accuracy,
                    'std_accuracy': result.std_accuracy,
                    'avg_error': result.avg_error,
                    'avg_loss': result.avg_loss
                }
            )

        return results

    @staticmethod
    def to_json(results: Dict) -> Dict:
        """Convert a result to its json representation."""

        # concatenate the results dict to a static one
        return dict({
            'version': {
                'schema': 1,
                'opset': 1
            }}, **results)

    # Methods for converting to proto.

    @staticmethod
    def _version_to_proto() -> Dict:
        return Version(
            schema=1,
            opset=1
        )

    @staticmethod
    def _runs_to_proto(results: Dict) -> Any:
        """converts results dictionary to protobuf InferenceRunsProto object"""

        # There are 4 fields in InferenceRunsProto, 3 are scalar
        inference_repeat = results['inference_repeat']
        is_partial = results['is_partial']
        time_elapsed = results['time_elapsed']

        # Create object with constructor specifying the scalar values only
        irp = InferenceRunsProto(
                inference_repeat=inference_repeat,
                is_partial=is_partial,
                time_elapsed=time_elapsed)

        # inference_results field is an array in protobuf and a list of
        #    dictionaries in the passed results

        # Build the InferenceResultsProto objects by appending each
        #    to the InferenceRunsProto object field
        i_results = results['inference_results']
        for result in i_results:
            irp.inference_results.append(       # pylint: disable=no-member
               InferenceResultsProto(
                   t_inference=result['t_inference'],
                   avg_accuracy=result['avg_accuracy'],
                   std_accuracy=result['std_accuracy'],
                   avg_error=result['avg_error'],
                   avg_loss=result['avg_loss']
               ))

        return irp



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Base drift compensation for inference."""

from typing import Tuple

from torch import Tensor
from torch.autograd import no_grad


class BaseDriftCompensation:
    """Base class for drift compensations."""

    def __init__(self) -> None:
        pass

    @no_grad()
    def init_baseline(self, forward_output: Tensor) -> Tuple[Tensor, Tensor]:
        """Initialize the base line for applying the compensation.

        Uses a all one tensor for read_out.

        Args:
            forward_output: forward output of the read out vector to compensate

        Returns:
            reference tensor readout
        """
        ref_value = self.readout(forward_output)

        return ref_value

    @no_grad()
    def get_readout_tensor(self, in_size: int) -> Tensor:
        """Return the read-out tensor.

        Called once during :meth:`~init_baseline`.
        """
        raise NotImplementedError

    @no_grad()
    def readout(self, out_tensor: Tensor) -> Tensor:
        """Implement the read out math."""
        raise NotImplementedError

    @no_grad()
    def apply(self, forward_output: Tensor, ref_value: Tensor) -> Tensor:
        """Read out the current value from the output of the forward
        pass and returns the drift compensation alpha scale."""
        current_value = self.readout(forward_output)
        ratio = ref_value / current_value

        return ratio



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: i_output_file.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import i_common_pb2 as i__common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13i_output_file.proto\x12\x11\x61ihwx.inferencing\x1a\x0ei_common.proto\"}\n\x15InferenceResultsProto\x12\x13\n\x0bt_inference\x18\x01 \x02(\x02\x12\x14\n\x0c\x61vg_accuracy\x18\x02 \x02(\x02\x12\x14\n\x0cstd_accuracy\x18\x03 \x02(\x02\x12\x11\n\tavg_error\x18\x04 \x02(\x02\x12\x10\n\x08\x61vg_loss\x18\x05 \x02(\x02\"\x9d\x01\n\x12InferenceRunsProto\x12\x18\n\x10inference_repeat\x18\x01 \x02(\x03\x12\x12\n\nis_partial\x18\x02 \x02(\x08\x12\x14\n\x0ctime_elapsed\x18\x03 \x02(\x02\x12\x43\n\x11inference_results\x18\x04 \x03(\x0b\x32(.aihwx.inferencing.InferenceResultsProto\"\x7f\n\x11InferencingOutput\x12+\n\x07version\x18\x01 \x02(\x0b\x32\x1a.aihwx.inferencing.Version\x12=\n\x0einference_runs\x18\x02 \x02(\x0b\x32%.aihwx.inferencing.InferenceRunsProto')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'i_output_file_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _INFERENCERESULTSPROTO._serialized_start=58
  _INFERENCERESULTSPROTO._serialized_end=183
  _INFERENCERUNSPROTO._serialized_start=186
  _INFERENCERUNSPROTO._serialized_end=343
  _INFERENCINGOUTPUT._serialized_start=345
  _INFERENCINGOUTPUT._serialized_end=472
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: output_file.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import onnx_common_pb2 as onnx__common__pb2
from . import common_pb2 as common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11output_file.proto\x12\x05\x61ihwx\x1a\x11onnx_common.proto\x1a\x0c\x63ommon.proto\"I\n\x10\x45pochInformation\x12\r\n\x05\x65poch\x18\x01 \x02(\x05\x12&\n\x07metrics\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"\xa3\x01\n\x0eTrainingOutput\x12\x1f\n\x07version\x18\x01 \x02(\x0b\x32\x0e.aihwx.Version\x12\x1f\n\x07network\x18\x02 \x02(\x0b\x32\x0e.aihwx.Network\x12\'\n\x06\x65pochs\x18\x03 \x03(\x0b\x32\x17.aihwx.EpochInformation\x12&\n\x07metrics\x18\x04 \x03(\x0b\x32\x15.aihwx.AttributeProto')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'output_file_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EPOCHINFORMATION._serialized_start=61
  _EPOCHINFORMATION._serialized_end=134
  _TRAININGOUTPUT._serialized_start=137
  _TRAININGOUTPUT._serialized_end=300
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Global drift compensation for inference."""

from torch.autograd import no_grad
from torch import abs as torch_abs
from torch import clamp, Tensor, eye

from aihwkit.inference.compensation.base import BaseDriftCompensation


class GlobalDriftCompensation(BaseDriftCompensation):
    """Global drift compensation.

    Uses a constant factor for compensating the drift.
    """

    @no_grad()
    def readout(self, out_tensor: Tensor) -> Tensor:
        """Read outs the mean abs."""
        return clamp(torch_abs(out_tensor).mean(), min=0.0001)

    @no_grad()
    def get_readout_tensor(self, in_size: int) -> Tensor:
        """Return the read-out tensor.

        Uses the set of one-hot vectors (eye).
        """
        return eye(in_size)

    def __str__(self) -> str:
        return '{}()'.format(self.__class__.__name__)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Visualization utilities.

This module includes plotting and visualization utilities for ``aihwkit``.
Using this module has extra dependencies that can be installed via the
extras mechanism::

    pip install aihwkit[visualization]
"""

# Allow untyped calls for `np.*`, as proper support for numpy typing requires
# 1.20+ (https://numpy.org/devdocs/reference/typing.html).
# mypy: disallow-untyped-calls=False

from copy import deepcopy
from typing import Any, Tuple, Union, Optional

import matplotlib.pyplot as plt
import numpy as np

from matplotlib import ticker
from matplotlib.figure import Figure
from numpy import ndarray
from torch import device as torch_device
from torch import eye, from_numpy, ones, stack

from aihwkit.exceptions import ConfigError
from aihwkit.simulator.configs import (
    SingleRPUConfig, UnitCellRPUConfig, InferenceRPUConfig
)
from aihwkit.simulator.configs.devices import PulsedDevice
from aihwkit.simulator.configs.compounds import UnitCell
from aihwkit.simulator.configs.utils import (
    BoundManagementType, IOParameters, NoiseManagementType, PulseType,
    UpdateParameters, WeightNoiseType
)
from aihwkit.simulator.tiles import AnalogTile, BaseTile, InferenceTile
from aihwkit.simulator.rpu_base import cuda
from aihwkit.inference.noise.base import BaseNoiseModel
from aihwkit.inference.noise.pcm import PCMLikeNoiseModel


def compute_pulse_response(
        analog_tile: BaseTile,
        direction: ndarray,
        use_forward: bool = False
) -> ndarray:
    """Compute the pulse response of a given device configuration.

    Args:
        analog_tile: Base tile used for computing the weight traces
        direction: numpy vector of directions to sequentially apply (-1 or 1)
        use_forward: Whether to use the (noisy) forward pass to read out the weights
            (otherwise returns exact weight value).

    Returns:
        An numpy array ``w_trace`` of dimensions ``len(direction) x out_size x in_size``
    """
    out_size = analog_tile.out_size
    in_size = analog_tile.in_size

    if analog_tile.is_cuda:
        device = torch_device('cuda')
    else:
        device = torch_device('cpu')

    total_iters = len(direction)
    w_trace = np.zeros((total_iters, out_size, in_size))
    in_vector = -ones(1, in_size, device=device)
    out_vector = ones(1, out_size, device=device)
    in_eye = eye(in_size, device=device)
    dir_tensor = from_numpy(direction).float().to(device)

    for i in range(total_iters):
        # Update the pulses.
        analog_tile.update(in_vector, out_vector * dir_tensor[i])

        if use_forward:
            # Save weights by using the forward pass (to get the short-term read noise).
            w_trace[i, :, :] = analog_tile.forward(in_eye).detach().cpu().numpy().T
        else:
            # Noise free.
            w_trace[i, :, :] = analog_tile.get_weights()[0].detach().cpu().numpy()

    return w_trace


def plot_pulse_response(
        analog_tile: BaseTile,
        direction: ndarray,
        use_forward: bool = False
) -> ndarray:
    """Plot the pulse response of a direction vector for each weight of
    the analog tile.

    Args:
        analog_tile: Base tile used for computing the weight traces
        direction: vector of directions to sequentially apply (-1 or 1)
        use_forward: Whether to use the (noisy) forward pass to read out the weights
            (otherwise returns exact weight value).

    Returns:
        w_trace: from :func:`compute_pulse_response`
    """
    w_trace = compute_pulse_response(analog_tile, direction, use_forward)

    plt.plot(w_trace.reshape(w_trace.shape[0], -1))
    if use_forward:
        plt.title(analog_tile.rpu_config.device.__class__.__name__)
    else:
        plt.title('{} (without cycle/read noise)'
                  .format(analog_tile.rpu_config.device.__class__.__name__))
    plt.ylabel('Weight [conductance]')
    plt.xlabel('Pulse number \\#')

    return w_trace


def compute_pulse_statistics(
        w_nodes: ndarray,
        w_trace: ndarray,
        direction: ndarray,
        up_direction: bool,
        smoothness: float = 0.5
) -> Tuple[ndarray, ndarray]:
    """Compute the statistics of the step trace from :func:`compute_pulse_response`.

    Args:
        w_nodes: weight range vector to estimate the step histogram
        w_trace: weight trace from :func:`compute_pulse_response`
        direction: direction vector used to generate the weight traces
        up_direction: whether and plot to compute the statistics for up or down direction
        smoothness: value for smoothing the estimation of the
            statistical step response curves

    Returns:
        Tuple of ``(dw_mean, dw_std)``.
    """
    # pylint: disable=too-many-locals

    def calc_mean_and_std(
            node: ndarray,
            w_values: ndarray,
            delta_w: ndarray,
            lam: float
    ) -> Tuple[ndarray, ndarray]:
        """Calculate the mean and std of a w location (node).

        Note:
            In case there are multiple trials then it also includes
            device-to-device variation.
        """
        alpha = np.exp(-0.5*(node - w_values)**2/lam**2)
        beta = alpha.sum(axis=0)
        alpha[:, beta < 0.1] = np.nan
        alpha /= np.expand_dims(beta, axis=0)  # type: ignore
        mean = np.sum(alpha*delta_w, axis=0)
        std = np.sqrt(np.sum(alpha*(delta_w - np.expand_dims(mean, axis=0))**2,  # type: ignore
                             axis=0))
        return (mean, std)

    # dw statistics.
    delta_w = np.diff(w_trace, axis=0)  # type: ignore
    w_trace_s = w_trace[:-1, :, :]
    if up_direction:
        msk = np.logical_and(direction[:-1] > 0, np.diff(direction) == 0)  # type: ignore
        w_values = w_trace_s[msk, :, :]
        delta_w_values = delta_w[msk, :, :]
    else:
        msk = np.logical_and(direction[:-1] < 0, np.diff(direction) == 0)  # type: ignore
        w_values = w_trace_s[msk, :, :]
        delta_w_values = delta_w[msk, :, :]

    dw_mean = np.zeros((len(w_nodes), w_trace.shape[1], w_trace.shape[2]))
    dw_std = np.zeros((len(w_nodes), w_trace.shape[1], w_trace.shape[2]))

    lam = (w_nodes[1]-w_nodes[0])/2*smoothness
    for i, node in enumerate(w_nodes):
        dw_mean[i, :, :], dw_std[i, :, :] = calc_mean_and_std(node, w_values, delta_w_values, lam)

    return dw_mean, dw_std


def plot_pulse_statistics(
        w_trace: ndarray,
        direction: ndarray,
        up_direction: bool,
        num_nodes: int = 100,
        smoothness: float = 0.5
) -> Tuple[ndarray, ndarray, ndarray]:
    """Plot the dG-G curve from a given weight trace and direction vector.

    Args:
        w_trace: weight trace from :func:`compute_pulse_response`
        direction: direction vector used to generate ``w_trace``
        up_direction: whether and plot to compute the statistics for up or down direction
        num_nodes: number of nodes for estimation of the step histogram
        smoothness: value for smoothing the estimation of the
            statistical step response curves

    Returns:
        A tuple (w_nodes, dw_mean, dw_std) from :func:`compute_pulse_statistics`
    """

    def errorbar_patch(x: ndarray, mean: ndarray, std: ndarray) -> None:
        """Plot a patchy error bar."""
        axis = plt.plot(x, mean)[0]
        plt.fill_between(x, mean - std, mean + std, edgecolor=None,
                         facecolor=axis.get_color(), alpha=0.5)

    # Compute statistics.
    w_nodes = np.linspace(w_trace.min(), w_trace.max(), num_nodes)
    dw_mean, dw_std = compute_pulse_statistics(w_nodes, w_trace, direction, up_direction,
                                               smoothness)

    n_traces = dw_mean.shape[1] * dw_mean.shape[2]

    for i in range(n_traces):
        errorbar_patch(w_nodes,
                       dw_mean.reshape(-1, n_traces)[:, i],
                       dw_std.reshape(-1, n_traces)[:, i])

    plt.xlabel('Weight $w$')
    plt.ylabel('Avg. step $\\Delta w$')
    if up_direction:
        plt.title('up-direction')
    else:
        plt.title('down-direction')

    return w_nodes, dw_mean, dw_std


def get_tile_for_plotting(
        rpu_config: Union[SingleRPUConfig, UnitCellRPUConfig],
        n_traces: int,
        use_cuda: bool = False,
        noise_free: bool = False
) -> BaseTile:
    """Return an analog tile for plotting the response curve.

    Args:
        rpu_config: RPU Configuration to use for plotting
        n_traces: Number of traces to plot
        use_cuda: Whether to use the CUDA implementation (if available)
        noise_free: Whether to turn-off cycle-to-cycle noises (if possible)

    Returns:
        Instantiated tile.
    """

    def set_noise_free(dev: Any) -> Any:
        if hasattr(dev, 'dw_min_std'):
            dev.dw_min_std = 0.0  # Noise free.

        if hasattr(dev, 'refresh_forward'):
            setattr(dev, 'refresh_forward', IOParameters(is_perfect=True))

        if hasattr(dev, 'refresh_update'):
            setattr(dev, 'refresh_update', UpdateParameters(pulse_type=PulseType.NONE))

        if hasattr(dev, 'transfer_forward'):
            setattr(dev, 'refresh_forward', IOParameters(is_perfect=True))

        if hasattr(dev, 'transfer_update'):
            setattr(dev, 'transfer_update', UpdateParameters(pulse_type=PulseType.NONE))

        if (hasattr(dev, 'write_noise_std') and
           getattr(dev, 'write_noise_std') > 0.0):
            # Just make very small to avoid hidden parameter mismatch.
            setattr(dev, 'write_noise_std', 1e-6)

    config = deepcopy(rpu_config)

    # Make sure we use single pulses for the overview.
    config.update.update_bl_management = False
    config.update.update_management = False
    config.update.desired_bl = 1

    if noise_free:
        config.forward.is_perfect = True

        set_noise_free(config.device)
        if hasattr(config.device, 'unit_cell_devices'):
            for dev in getattr(config.device, 'unit_cell_devices'):
                set_noise_free(dev)
        if hasattr(config.device, 'device'):
            set_noise_free(getattr(config.device, 'device'))

    analog_tile = AnalogTile(n_traces, 1, config)  # type: BaseTile
    analog_tile.set_learning_rate(1)
    w_min = getattr(config.device.as_bindings(), 'w_min', -1.0)

    weights = w_min * ones((n_traces, 1))
    analog_tile.set_weights(weights)

    if use_cuda and cuda.is_compiled():
        return analog_tile.cuda()
    return analog_tile


def estimate_n_steps(rpu_config: Union[SingleRPUConfig, UnitCellRPUConfig]) -> int:
    """Estimate the n_steps.

    Note:
        The estimate of the number of update pulses needed to drive
        from smallest to largest conductance. The estimation just
        assumes linear behavior, thus only be a rough estimate for
        non-linear response curves.

    Args:
        rpu_config: RPU Configuration to use for plotting

    Returns:
        Guessed number of steps

    Raises:
        ConfigError: If rpu_config.device does not have the w_min
            attribute (which is only ensured for
            :class:`~aihwkit.simulator.configs.devices.PulseDevice`)
    """
    if not isinstance(rpu_config, SingleRPUConfig):
        return 1000

    device_binding = rpu_config.device.as_bindings()

    if not hasattr(device_binding, 'w_min'):
        raise ConfigError('n_step estimation only for PulsedDevice. ' +
                          'Provide n_step explicitly.')

    weight_granularity = device_binding.calc_weight_granularity()
    w_min = device_binding.w_min
    w_max = device_binding.w_max

    n_steps = int(np.round((w_max - w_min) / weight_granularity))  # type: ignore
    return n_steps


def plot_response_overview(
        rpu_config: Union[SingleRPUConfig, UnitCellRPUConfig],
        n_loops: int = 5,
        n_steps: Optional[int] = None,
        n_traces: int = 5,
        use_cuda: bool = False,
        smoothness: float = 0.1
) -> None:
    """Plot the step response and statistics of a given device configuration.

    Args:
        rpu_config: RPU Configuration to use for plotting
        n_loops: How many hyper-cycles (up/down pulse sequences) to plot
        n_steps: Number of up/down steps per cycle. If not given, will
            be tried to be estimated (only for ``PulsedDevice``
            possible otherwise defaults to 1000 if ``n_steps=None``).
        n_traces: Number of traces to plot
        use_cuda: Whether to use the CUDA implementation (if available)
        smoothness: value for smoothing the estimation of the
            statistical step response curves
    """
    if n_steps is None:
        n_steps = estimate_n_steps(rpu_config)

    total_iters = min(max(n_loops*2*n_steps, 1000), max(50000, 2*n_steps))
    direction = np.sign(np.sin(np.pi*(np.arange(total_iters)+1)/n_steps))

    plt.clf()

    # 1. Noisy tile.
    analog_tile = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=False)
    plt.subplot(3, 1, 1)
    plot_pulse_response(analog_tile, direction, use_forward=True)

    # 2. Noise-free tile.
    analog_tile_noise_free = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=True)
    analog_tile_noise_free.set_hidden_parameters(analog_tile.get_hidden_parameters())

    plt.subplot(3, 1, 2)
    w_trace = plot_pulse_response(analog_tile_noise_free, direction, use_forward=False)

    num_nodes = min(n_steps, 100)
    # 3. Plot up statistics.
    plt.subplot(3, 2, 5)
    plot_pulse_statistics(w_trace, direction, True, num_nodes, smoothness)

    # 4. Plot down statistics.
    plt.subplot(3, 2, 6)
    plot_pulse_statistics(w_trace, direction, False, num_nodes, smoothness)

    plt.tight_layout()


def plot_device(device: Union[PulsedDevice, UnitCell], w_noise: float = 0.0, **kwargs: Any) -> None:
    """Plot the step response figure for a given device (preset).

    Note:
        It will use an amount of read weight noise ``w_noise`` for
        reading the weights.

    Args:
        device: PulsedDevice parameters
        w_noise: Weight noise standard deviation during read
        kwargs: for other parameters, see :func:`plot_response_overview`
    """
    plt.figure(figsize=[7, 7])
    # To simulate some weight read noise.
    io_pars = IOParameters(out_noise=0.0,    # no out noise
                           w_noise=w_noise,  # quite low
                           inp_res=-1.,      # turn off DAC
                           out_bound=100.,   # not limiting
                           out_res=-1.,      # turn off ADC
                           bound_management=BoundManagementType.NONE,
                           noise_management=NoiseManagementType.NONE,
                           w_noise_type=WeightNoiseType.ADDITIVE_CONSTANT)

    if isinstance(device, PulsedDevice):
        plot_response_overview(SingleRPUConfig(device=device, forward=io_pars), **kwargs)
    else:
        plot_response_overview(UnitCellRPUConfig(device=device, forward=io_pars), **kwargs)


def plot_device_compact(
        device: Union[PulsedDevice, UnitCell],
        w_noise: float = 0.0,
        n_steps: Optional[int] = None,
        n_traces: int = 3,
        use_cuda: bool = False,
) -> Figure:
    """Plot a compact step response figure for a given device (preset).

    Note:
        It will use an amount of read weight noise ``w_noise`` for
        reading the weights.

    Args:
        device: ``PulsedDevice`` or ``UnitCell`` parameters
        w_noise: Weight noise standard deviation during read
        n_steps: Number of up/down steps per cycle. If not given, will
            be tried to be estimated (only for ``PulsedDevice``
            possible otherwise defaults to 1000 if ``n_steps=None``).
        n_traces: Number of traces to plot (for device-to-device variation)
        use_cuda: Whether to use CUDA for the computation

    Returns:
        the compact step response figure.
    """
    # pylint: disable=too-many-locals,too-many-statements
    def get_rpu_config(device: Union[PulsedDevice, UnitCell], io_pars: IOParameters) \
            -> Union[SingleRPUConfig, UnitCellRPUConfig]:
        if isinstance(device, PulsedDevice):
            return SingleRPUConfig(device=device, forward=io_pars)
        return UnitCellRPUConfig(device=device, forward=io_pars)

    figure = plt.figure(figsize=[12, 4])

    # To simulate some weight read noise.
    io_pars = IOParameters(out_noise=0.0,    # no out noise
                           w_noise=w_noise,  # quite low
                           inp_res=-1.,      # turn off DAC
                           out_bound=100.,   # not limiting
                           out_res=-1.,      # turn off ADC
                           bound_management=BoundManagementType.NONE,
                           noise_management=NoiseManagementType.NONE,
                           w_noise_type=WeightNoiseType.ADDITIVE_CONSTANT)

    rpu_config = get_rpu_config(device, io_pars)

    if n_steps is None:
        n_steps = estimate_n_steps(rpu_config)

    # Noisy tile response curves.
    n_loops = 2
    total_iters = n_loops*2*n_steps
    direction = np.sign(np.sin(np.pi*(np.arange(total_iters)+1)/n_steps))

    analog_tile = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=False)
    w_trace = compute_pulse_response(analog_tile, direction, use_forward=True)\
        .reshape(-1, n_traces)
    axis = figure.add_subplot(1, 1, 1)
    axis.plot(w_trace, linewidth=1)
    axis.set_title(analog_tile.rpu_config.device.__class__.__name__)
    axis.set_xlabel('Pulse number \\#')
    limit = np.abs(w_trace).max()*1.2
    axis.set_ylim(-limit, limit)
    axis.set_xlim(0, total_iters-1)
    axis.xaxis.set_major_formatter(ticker.ScalarFormatter(useMathText=True))

    # Noise-free tile for statistics.
    n_loops = 1
    total_iters = min(max(n_loops*2*n_steps, 1000), max(50000, 2*n_steps))
    direction = np.sign(np.sin(np.pi*(np.arange(total_iters)+1)/n_steps))

    analog_tile_noise_free = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=True)
    analog_tile_noise_free.set_hidden_parameters(analog_tile.get_hidden_parameters())

    w_trace = compute_pulse_response(analog_tile_noise_free, direction, False)

    # Compute statistics.
    num_nodes = min(n_steps, 100)
    w_nodes = np.linspace(w_trace.min(), w_trace.max(), num_nodes)

    dw_mean_up = compute_pulse_statistics(w_nodes, w_trace, direction, True)[0]\
        .reshape(-1, n_traces)
    dw_mean_down = compute_pulse_statistics(w_nodes, w_trace, direction, False)[0]\
        .reshape(-1, n_traces)

    # Plot mean up statistics.
    pos = axis.get_position().bounds
    space = 0.1
    gap = 0.01
    axis.set_position([pos[0] + gap + space, pos[1], pos[2] - 2*gap - 2*space, pos[3]])
    axis.set_yticks([])

    axis_left = figure.add_axes([pos[0], pos[1], space, pos[3]])
    dw_mean_up = dw_mean_up.reshape(-1, n_traces)
    for i in range(n_traces):
        axis_left.plot(dw_mean_up[:, i], w_nodes)

    axis_left.set_position([pos[0], pos[1], space, pos[3]])
    axis_left.set_xlabel('Up pulse size')
    axis_left.set_ylabel('Weight \n [conductance]')
    axis_left.set_ylim(-limit, limit)

    # Plot mean down statistics.
    axis_right = figure.add_axes([pos[0] + pos[2] - space, pos[1], space, pos[3]])
    dw_mean_down = dw_mean_down.reshape(-1, n_traces)
    for i in range(n_traces):
        axis_right.plot(np.abs(dw_mean_down[:, i]), w_nodes)

    axis_right.set_yticks([])
    axis_right.set_xlabel('Down pulse size')
    axis_right.set_ylim(-limit, limit)

    # Set xlim's.
    limit = np.maximum(np.nanmax(np.abs(dw_mean_down)),  # type: ignore
                       np.nanmax(np.abs(dw_mean_up))) * 1.2  # type: ignore
    axis_left.set_xlim(0.0, limit)
    axis_right.set_xlim(0.0, limit)

    return figure


def plot_device_symmetry(
        device: PulsedDevice,
        w_noise: float = 0.0,
        n_pulses: int = 10000,
        n_traces: int = 3,
        use_cuda: bool = False,
        w_init: float = 1.0,
) -> None:
    """Plot the response figure for a given device (preset).

    It will show the response to alternating up down pulses.

    Note:
        It will use an amount of read weight noise ``w_noise`` for
        reading the weights.

    Args:
        device: PulsedDevice parameters
        n_pulses: total number of pulses
        w_noise: Weight noise standard deviation during read
        n_traces: Number of device traces
        use_cuda: Whether to use CUDA,
        w_init: Initial value of the weights
    """
    plt.figure(figsize=[10, 5])

    io_pars = IOParameters(out_noise=0.0,    # no out noise
                           w_noise=w_noise,  # quite low
                           inp_res=-1.,      # turn off DAC
                           out_bound=100.,   # not limiting
                           out_res=-1.,      # turn off ADC
                           bound_management=BoundManagementType.NONE,
                           noise_management=NoiseManagementType.NONE,
                           w_noise_type=WeightNoiseType.ADDITIVE_CONSTANT)

    rpu_config = SingleRPUConfig(device=device, forward=io_pars)

    direction = np.sign(np.cos(np.pi*np.arange(n_pulses)))
    plt.clf()

    analog_tile = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=False)
    weights = w_init*ones((n_traces, 1))
    analog_tile.set_weights(weights)

    plot_pulse_response(analog_tile, direction, use_forward=False)
    plt.ylim([-1, 1])
    plt.grid(True)


def plot_weight_drift(noise_model: Optional[BaseNoiseModel] = None,
                      t_inference_list: Optional[ndarray] = None,
                      w_inits: Optional[ndarray] = None,
                      n_repeats: int = 25) -> None:
    """Plots the weight drift behavior of a given noise model over time.

    Args:
        noise_model: Noise model of derived from
            :class:`~aihwkit.simulator.noise_models.BaseNoiseModel`
        t_inference_list: Numpy array of times of inference after
            programming at time 0 (in seconds)
        w_inits: Numpy array of target weights to program
        n_repeats: How many repeats to estimate the standard deviation
    """
    # pylint: disable=too-many-locals

    plt.figure(figsize=[10, 5])
    if noise_model is None:
        noise_model = PCMLikeNoiseModel()
    if t_inference_list is None:
        t_inference_list = np.logspace(0., 7.0, 15)
    if w_inits is None:
        w_inits = np.linspace(-1., 1., 9)

    rpu_config = InferenceRPUConfig(noise_model=noise_model, drift_compensation=None)

    weights = w_inits.flatten()
    weights.sort()
    weights = np.tile(weights, [n_repeats, 1])  # type: ignore

    analog_tile = InferenceTile(weights.shape[0], weights.shape[1], rpu_config)
    analog_tile.set_weights(from_numpy(weights))
    analog_tile.program_weights()
    programmed_weights, _ = analog_tile.get_weights()

    m_list = []
    s_list = []

    for t_inference in t_inference_list:
        analog_tile.drift_weights(t_inference=t_inference)
        noisy_weights, _ = analog_tile.get_weights()

        m_list.append(noisy_weights.mean(axis=0))
        s_list.append((programmed_weights - noisy_weights).std(axis=0))

    m_array = stack(m_list, axis=1)
    s_array = stack(s_list, axis=1)

    for i in range(w_inits.size):
        curve = plt.plot(t_inference_list, m_array[i])

        plt.fill_between(t_inference_list,
                         m_array[i] - s_array[i],
                         m_array[i] + s_array[i],
                         edgecolor=None, linewidth=0, alpha=0.3, antialiased=True,
                         facecolor=curve[0].get_color())

    plt.gca().set_xscale('log')
    plt.xlabel('Time after programming [sec]')
    plt.ylabel('Weight value [norm. units]')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: input_file.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10input_file.proto\x12\x05\x61ihwx\x1a\x0c\x63ommon.proto\"\\\n\x07\x44\x61taset\x12\x12\n\ndataset_id\x18\x01 \x02(\t\x12)\n\ntransforms\x18\x02 \x03(\x0b\x32\x15.aihwx.TransformProto\x12\x12\n\nbatch_size\x18\x03 \x01(\x05\"\x9f\x01\n\x08Training\x12(\n\toptimizer\x18\x01 \x02(\x0b\x32\x15.aihwx.OptimizerProto\x12(\n\tscheduler\x18\x02 \x01(\x0b\x32\x15.aihwx.SchedulerProto\x12/\n\rloss_function\x18\x03 \x01(\x0b\x32\x18.aihwx.LossFunctionProto\x12\x0e\n\x06\x65pochs\x18\x04 \x01(\x05\"\x95\x01\n\rTrainingInput\x12\x1f\n\x07version\x18\x01 \x02(\x0b\x32\x0e.aihwx.Version\x12\x1f\n\x07\x64\x61taset\x18\x02 \x02(\x0b\x32\x0e.aihwx.Dataset\x12\x1f\n\x07network\x18\x03 \x02(\x0b\x32\x0e.aihwx.Network\x12!\n\x08training\x18\x04 \x02(\x0b\x32\x0f.aihwx.Training')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'input_file_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DATASET._serialized_start=41
  _DATASET._serialized_end=133
  _TRAINING._serialized_start=136
  _TRAINING._serialized_end=295
  _TRAININGINPUT._serialized_start=298
  _TRAININGINPUT._serialized_end=447
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Phenomenological noise model for inference."""

from copy import deepcopy
from typing import List, Optional

from numpy import log as numpy_log
from numpy import sqrt
from torch import abs as torch_abs
from torch import randn_like, Tensor
from torch.autograd import no_grad

from aihwkit.inference.noise.base import BaseNoiseModel
from aihwkit.inference.converter.base import BaseConductanceConverter
from aihwkit.inference.converter.conductance import SinglePairConductanceConverter


class StateIndependentNoiseModel(BaseNoiseModel):  # pylint: disable=too-many-instance-attributes
    r"""Standard noise model that has a non-conductance dependent drift and
    multiplicative read (1/f) noise.

    Programming noise is state-independent by default, however, it can
    be made conductance dependent, since the expected programming
    noise strength is modeled with a second-order polynomial in
    general.

   **Programming noise** is thus given by:

    .. math::

         \sigma_text{programming noise}=\gamma\,\left(c_0 + c_1 \frac{g_T}{g_\text{max}} +
              + c_2 \frac{g_T^2}{g_\text{max}^2}\right)

    where :math:`\gamma` is a additional convenience scale and :math:`g_T`
    is the target conductance established from the given
    ``g_converter`` from the weight matrix.  The default programming
    noise is constant (state independent): :math:`c_0=0.2\mu\mathrm{S}`
    and other coefficient set to :math:`0.0`.

    **Drift** is for each device is computed as

    .. math::

        g_\text{drift}(t) = g_\text{prog}(t / t_0) ^{- \nu}

    with the drift coefficient determined at the beginning for each
    device with

    .. math::

        \nu= \zeta\, |\nu_\text{mean} + \nu_\text{std}\xi|_+

    where :math:`\xi` is a Gaussian random number and
    :math:`|\cdot|_+` rectifies negative value to zero. :math:`\zeta`
    is an additional drift scale.

    **Read noise** is given by

    .. math::

        \sigma_\text{read} = \rho \frac{g_\text{drift}(t)}{g_\text{max}} \sqrt{\log\left(\frac{t
        + t_\text{read}}{2 t_\text{read}}\right)}

    This :math:`\sigma_\text{read}` is then used to add Gaussian noise
    of this magnitude to the drifted conductance. The read noise scale
    :math:`\rho` can be used to scale the read noise.

    Args:
        g_converter: instantiated class of the conductance converter
            (defaults to single pair)

        g_max: In :math:`\mu S`, the maximal conductance, ie the value
            the absolute max of the weights will be mapped to.

        prog_coeff: programming polynomial coefficients :math:`c_i` in
            :math:`\mu S`. Default is constant :math:`c_0=0.2` and
            other coefficient set to 0.0.

        prog_noise_scale: scale :math:\gamma: for the programming noise

        drift_nu_mean: mean :math:`\nu_\text{mean}` of power-law drift
            coefficient (:math:`\nu`) (before ``drift_scale``
            :math:`\zeta` is applied).

        drift_nu_std: device-to-device variability
            :math:`\nu_\text{std}` of the power-law drift coefficient
            (before ``drift_scale`` is applied)

        drift_scale: additional scale :math:`\zeta` applied to all
            drawn drift coefficients

        t_0: parameter of the drift (first reading time), see above.

            Note:
                The ``t_inference`` is relative to this time ``t0``
                e.g. ``t_inference`` counts from the completion of the
                programming of a device.

        read_noise_scale: scale :math:`\rho` for scaling the read and
            accumulated noise :math:`1/f`.

        t_read: parameter of the :math:`1/f` noise (in seconds)

    """

    def __init__(  # pylint: disable=too-many-arguments
            self,
            g_converter: Optional[BaseConductanceConverter] = None,
            g_max: Optional[float] = None,
            prog_coeff: Optional[List[float]] = None,
            prog_noise_scale: float = 1.0,
            drift_nu_mean: float = 0.1,
            drift_nu_std: float = 0.05,
            drift_scale: float = 1.0,
            t_0: float = 20.0,
            read_noise_scale: float = 1.0,
            t_read: float = 250.0e-9,
    ):
        g_converter = deepcopy(g_converter) or SinglePairConductanceConverter(g_max=g_max)
        super().__init__(g_converter)

        self.g_max = getattr(self.g_converter, 'g_max', g_max)

        if self.g_max is None:
            raise ValueError('g_max cannot be established from g_converter')

        self.prog_coeff = [0.2, 0.0, 0.0] if prog_coeff is None else prog_coeff
        self.prog_noise_scale = prog_noise_scale
        self.drift_nu_mean = drift_nu_mean
        self.drift_nu_std = drift_nu_std
        self.drift_scale = drift_scale
        self.t_0 = t_0
        self.read_noise_scale = read_noise_scale
        self.t_read = t_read

    def __str__(self) -> str:
        ret = self.__class__.__name__ + '('
        for key in ['g_converter', 'prog_coeff', 'prog_noise_scale',
                    'drift_nu_mean', 'drift_nu_std', 'drift_scale',
                    't_0', 'read_noise_scale', 't_0']:
            ret += key + '={}, '.format(self.__dict__[key])
        ret = ret[:-2] + ')'
        return ret

    @no_grad()
    def apply_programming_noise_to_conductance(self, g_target: Tensor) -> Tensor:
        """Apply programming noise to a target conductance Tensor.

        Programming noise with additive Gaussian noise with
        conductance dependency of the variance given by a 2-degree
        polynomial.
        """
        mat = 1
        sig_prog = self.prog_coeff[0]
        for coeff in self.prog_coeff[1:]:
            mat *= g_target / self.g_max
            sig_prog += mat * coeff

        g_prog = g_target + self.prog_noise_scale * sig_prog * randn_like(g_target)
        g_prog.clamp_(min=0.0)  # no negative conductances allowed
        return g_prog

    @no_grad()
    def generate_drift_coefficients(self, g_target: Tensor) -> Tensor:
        """Return drift coefficients ``nu``."""

        mu_drift = self.drift_nu_mean
        sig_drift = self.drift_nu_std
        nu_drift = torch_abs(mu_drift + sig_drift * randn_like(g_target)).clamp(min=0.0)
        return nu_drift * self.drift_scale

    @no_grad()
    def apply_drift_noise_to_conductance(
            self,
            g_prog: Tensor,
            nu_drift: Tensor,
            t_inference: float
    ) -> Tensor:
        """Apply the noise and drift up to the assumed inference time
        point."""
        t = t_inference + self.t_0

        # drift
        if t > self.t_0:
            g_drift = g_prog * ((t / self.t_0) ** (- nu_drift))
        else:
            g_drift = g_prog

        # expected accumulated 1/f noise since start of programming at t=0
        if t > 0:
            sig_noise = sqrt(numpy_log((t + self.t_read) / (2 * self.t_read)))
            g_final = g_drift + torch_abs(g_drift / self.g_max) * self.read_noise_scale \
                * sig_noise * randn_like(g_drift)
        else:
            g_final = g_prog

        return g_final.clamp(min=0.0)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Basic inferencing Experiment."""

# pylint: disable=too-many-locals

from typing import Any, Dict, Tuple, Type, Optional
from os import path, mkdir
from copy import deepcopy
from requests import get as requests_get
from numpy import ndarray, array, logspace, log10, zeros, concatenate


from torch import device as torch_device, max as torch_max, Tensor
from torch import load
from torch.nn import Module, CrossEntropyLoss
from torch.nn.modules.loss import _Loss
from torch.utils.data import DataLoader, Dataset, Subset
from torchvision.datasets import FashionMNIST, SVHN
from torchvision.transforms import Compose, Normalize, ToTensor

from aihwkit.experiments.experiments.base import Experiment, Signals
from aihwkit.nn.modules.base import AnalogModuleBase


WEIGHT_TEMPLATE_URL = "https://github.com/IBM-AI-Hardware-Center/Composer/raw/main/"


def download(url: str, destination: str) -> None:
    """Helper for downloading a file from url"""
    response = requests_get(url, timeout=30.0)
    with open(destination, "wb") as file_:
        file_.write(response.content)


class BasicInferencing(Experiment):
    """Experiment for inferencing a neural network.

    ``Experiment`` that represents inferencing a neural network using a basic
    inferencing loop.

    This class contains:

    * the data needed for an experiment. The recommended way of setting this
      data is via the arguments of the constructor. Additionally, some of the
      items have getters that are used by the ``Workers`` that execute the
      experiments and by the inferencing loop.
    * the inferencing algorithm, with the main entry point being ``train()``.

    Note:
        When executing a ``BasicInferencing`` in the cloud, additional constraints
        are applied to the data. For example, the model is restricted to
        sequential layers of specific types; the dataset choices are limited,
        etc. Please check the ``CloudRunner`` documentation.
    """

    def __init__(
            self,
            dataset: Type[Dataset],
            model: Module,
            batch_size: int = 10,
            loss_function: type = CrossEntropyLoss,
            weight_template_id: str = "",
            inference_repeats: int = 2,
            inference_time: int = 86400,
            remap_weights: bool = True
    ):
        """Create a new ``BasicInferencing``.

        Args:
            dataset: the dataset class to be used.
            model: the neural network to use for inferencing.
            batch_size: the batch size used for inferencing.
            loss_function: the loss function used in the neural network.
            weight_template_id: weights and biases of the trained neural network.
            inference_repeats: the number of times running the inference.
            inference_time: the time span between programming the chip and performing the inference.
            remap_weights:  whether to remap the weights
        """
        self.dataset = dataset
        self.model = model
        self.batch_size = batch_size
        self.loss_function = loss_function
        self.inference_repeats = inference_repeats
        self.inference_time = inference_time
        self.weight_template_id = weight_template_id
        self.remap_weights = remap_weights

        super().__init__()

    def get_dataset_arguments(self, dataset: type) -> Tuple[Dict, Dict]:
        """Return the dataset constructor arguments for specifying subset."""
        if dataset in (SVHN, ):
            return {'split': 'train'}, {'split': 'test'}
        return {'train': True}, {'train': False}

    def get_dataset_transform(self, dataset: type) -> Any:
        """Return the dataset transform."""
        # Normalize supported datasets.
        if dataset == FashionMNIST:
            # mean = Tensor([0.2860])
            # std_dev = Tensor([0.3205])
            # transform = Compose([ToTensor(), Normalize(mean, std_dev)])
            # Note: I removed the normalize step to match up with
            # the steps that were used by Fabio to generate the weight file.
            transform = Compose([ToTensor()])
        elif dataset == SVHN:
            mean = Tensor([0.4377, 0.4438, 0.4728])
            std_dev = Tensor([0.1980, 0.2010, 0.1970])
            transform = Compose([ToTensor(), Normalize(mean, std_dev)])
        else:
            transform = Compose([ToTensor()])

        return transform

    def get_data_loader(
            self,
            dataset: type,
            batch_size: int,
            max_elements: int = 0,
            dataset_root: str = '/tmp/datasets'
    ) -> DataLoader:
        """Return `DataLoaders` for the selected dataset.

        Args:
            dataset: the dataset class to be used.
            batch_size: the batch size used for inferencing.
            max_elements: the maximum number of elements of the dataset
                to be used. If ``0``, the full dataset is used.
            dataset_root: the path to the folder where the files from the
                dataset are stored.

        Returns:
            A tuple with the inferencing and validation loaders.
        """
        # Create the sets and the loaders.
        _, test_args = self.get_dataset_arguments(dataset)
        transform = self.get_dataset_transform(dataset)

        # Create the validation set
        validation_set = dataset(dataset_root, transform=transform, **test_args)

        if max_elements > 0:
            validation_set = Subset(validation_set, range(max_elements))

        validation_loader = DataLoader(validation_set, batch_size=batch_size, shuffle=False)

        return validation_loader

    def get_model(
            self,
            weight_template_id: str,
            device: torch_device
    ) -> Module:
        """Get a copy of the set-up model (with load the weights and biases)
        from the original experiment model.

        Args:
            weight_template_id: location/index for the file
                that contains the state_dicts for the model.
            device: the torch device used for the model.

        Returns:
            a copied model with loaded weights and biases

        """
        model = deepcopy(self.model)

        if weight_template_id != "":
            if weight_template_id[0:1] == "." or weight_template_id[0:1] == '/':
                # This is the case where it is a local file
                template_path = weight_template_id
            else:
                template_dir = "/tmp/weight_templates"
                if weight_template_id.startswith('http'):
                    template_url = weight_template_id
                else:
                    # print('weights_template_id: ', weight_template_id)
                    template_path = template_dir + "/" + weight_template_id + ".pth"
                    template_url = WEIGHT_TEMPLATE_URL + weight_template_id + ".pth"
                # check if the file exists
                if not path.exists(template_dir):
                    mkdir(template_dir)
                if not path.exists(template_path):
                    download(template_url, template_path)

            # print('template_path: ', template_path)
            if path.exists(template_path):
                model.load_state_dict(load(template_path, map_location=device),
                                      load_rpu_config=False)
            else:
                print('Checkpoint file: ', template_path, ' does not exist.')

            if self.remap_weights:
                for module in model.analog_modules():
                    module.remap_weights()

        return model.to(device)

    def inferencing_step(
            self,
            validation_loader: DataLoader,
            model: Module,
            loss_function: _Loss,
            t_inference_list: list,
            device: torch_device
    ) -> Tuple[ndarray, ndarray, ndarray]:
        """Run a single inferencing.

        Args:
            validation_loader: the data loader for the inferencing data.
            model: the neural network to be trained.
            loss_function: the loss function used for inferencing.
            t_inference_list: list of t_inferences.
            device: the torch device used for the model.

        Return:
            Tuple of ndarray of inference accuracy, error and loss.
        """

        # Set the mode mode to eval mode.
        model.eval()
        # Reset the program analog weights.
        model.program_analog_weights()

        n_inference = len(t_inference_list)
        infer_accuracy = zeros(n_inference)
        infer_error = zeros(n_inference)
        infer_loss = zeros(n_inference)

        # Simulation of inference pass at different times after training.
        # Go through the generated list
        for idx, t_inference in enumerate(t_inference_list):
            # Set the drift_analog_weights
            model.drift_analog_weights(t_inference)

            # Initialize variables as needed.
            predicted_ok = 0
            total_images = 0
            total_loss = 0

            # Go through the images in the validation dataset
            for images, labels in validation_loader:
                # Load the images and labels into the memory of the device
                images = images.to(device)
                labels = labels.to(device)
                # Do prediction for the images using the model.
                predict = model(images)
                # Calculate the loss
                loss = loss_function(predict, labels)

                # Cummulate the loss to total_loss
                n_images = images.size(0)
                total_images += n_images
                total_loss += loss.item() * n_images

                _, predicted = torch_max(predict.data, 1)
                predicted_ok += (predicted == labels).sum().item()

            # Save the information in the np arrays and return
            accuracy_post = predicted_ok / total_images * 100.
            infer_accuracy[idx] = accuracy_post
            infer_error[idx] = 100. - accuracy_post
            infer_loss[idx] = total_loss / total_images

        return infer_accuracy, infer_error, infer_loss

    def inference(
            self,
            validation_loader: DataLoader,
            model: Module,
            loss_function: _Loss,
            inference_repeats: int,
            inference_time: int,
            device: torch_device,
            n_inference_times: int = 10,
    ) -> Dict:
        """Run the inferencing loop.

        Args:
            validation_loader: the data loader for the validation data.
            model: the neural network to be trained.
            loss_function: the loss function used for inferencing.
            inference_repeats: the number of times to repeat the process
                zof programming and drifting.
            inference_time: the time span between programming the chip and performing the inference.
            device: the torch device used for the model.
            n_inference_times: how many inference times (log-spaced)

        Returns:
            A list of the metrics for each epoch.
        """

        # Move the model to the device if needed.
        if device:
            model = model.to(device)

        # Create the t_inference_list using inference_time.
        # Generate the 9 values between 0 and the inference time using log10
        t_inference_list = [0.0] + logspace(0, log10(float(inference_time)),
                                            n_inference_times - 1).tolist()
        repeat_results = {}
        accuracy_array = array([])
        error_array = array([])
        loss_array = array([])

        for repeat in range(inference_repeats):
            self._call_hook(Signals.INFERENCE_REPEAT_START, repeat)
            infer_accuracy, infer_error, infer_loss = self.inferencing_step(
                validation_loader, model, loss_function, t_inference_list, device)

            # Save the info
            accuracy_array = concatenate([accuracy_array, infer_accuracy])  # type: ignore
            error_array = concatenate([error_array, infer_error])  # type: ignore
            loss_array = concatenate([loss_array, infer_loss])  # type: ignore

            # call the metric hook function with the average information
            # to write out the partial result to standard out.
            shape = (repeat + 1, n_inference_times)
            repeat_results = self._call_hook(Signals.INFERENCE_REPEAT_END,
                                             array(t_inference_list),
                                             accuracy_array.reshape(shape).mean(axis=0),
                                             accuracy_array.reshape(shape).std(axis=0),
                                             error_array.reshape(shape).mean(axis=0),
                                             loss_array.reshape(shape).mean(axis=0),
                                             self.inference_repeats)
        return deepcopy(repeat_results)

    def _print_rpu_fields(self, model: Module) -> None:
        """Print the Inference RPU Config fields"""

        print('\n>>> inferenceworker.py: STARTING _print_rpu_fields() ')

        for name, module in model.named_modules():
            if not isinstance(module, AnalogModuleBase):
                continue

            print(f'RPUConfig of module {name}:')
            tile = next(module.analog_tiles())
            print(tile.rpu_config)
            print(tile.tile)
            print('-------------')

        print('\n>>> inferenceworker.py: ENDING _print_rpu_fields() ')

    def run(self, max_elements: int = 0,
            dataset_root: str = '/tmp/data',
            device: Optional[torch_device] = None,
            ) -> Dict:
        """ Sets up the internal model and runs the inference.

        Results are returned and the internal model is updated.
        """

        # Build the objects needed for inferencing.
        # Get valication dataset
        validation_loader = self.get_data_loader(
            self.dataset, self.batch_size,
            max_elements=max_elements,
            dataset_root=dataset_root
        )

        # Load the weights and biases to the model.
        # Assumption: the model already includes the customer-specified InferenceRPUConfig.
        model = self.get_model(self.weight_template_id, device)

        self._print_rpu_fields(model)

        # Invoke the inference step
        result = self.inference(validation_loader,
                                model,
                                self.loss_function(),
                                self.inference_repeats,
                                self.inference_time,
                                device)
        self.model = model  # update the stored model with the trained one
        return result

    def __str__(self) -> str:
        """Return a string representation of a BasicInferencing experiment."""
        return ('{}(dataset={}, batch_size={}, loss_function={}, inference_repeats={}, '
                'inference_time={}, model={})'.format(
                    self.__class__.__name__,
                    getattr(self.dataset, '__name__', self.dataset),
                    self.batch_size,
                    getattr(self.loss_function, '__name__', self.loss_function),
                    self.inference_repeats,
                    self.inference_time,
                    self.model
                ))



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Analog RNN modules. """

import warnings
import math
from typing import Any, List, Optional, Tuple, Type, Callable
from torch import Tensor, jit
from torch.nn import Dropout, ModuleList, init

from aihwkit.nn.modules.container import AnalogSequential
from aihwkit.nn.modules.rnn.layers import AnalogRNNLayer, AnalogBidirRNNLayer
from aihwkit.nn.modules.base import AnalogModuleBase, RPUConfigAlias


class ModularRNN(AnalogSequential):
    """Helper class to create a Modular RNN

    Args:
        num_layers: number of serially connected RNN layers
        layer: RNN layer type (e.g. AnalogLSTMLayer)
        dropout: dropout applied to output of all RNN layers except last
        first_layer_args: RNNCell type, input_size, hidden_size, rpu_config, etc.
        other_layer_args: RNNCell type, hidden_size, hidden_size, rpu_config, etc.
    """
    # pylint: disable=abstract-method

    # Necessary for iterating through self.layers and dropout support
    __constants__ = ['layers', 'num_layers']

    def __init__(
            self,
            num_layers: int,
            layer: Type,
            dropout: float,
            first_layer_args: Any,
            other_layer_args: Any):
        super().__init__()
        self.layers = self.init_stacked_analog_lstm(num_layers, layer, first_layer_args,
                                                    other_layer_args)

        # Introduce a Dropout layer on the outputs of each RNN layer except
        # the last layer.
        self.num_layers = num_layers
        if num_layers == 1 and dropout > 0:
            warnings.warn('dropout lstm adds dropout layers after all but last '
                          'recurrent layer, it expects num_layers greater than '
                          '1, but got num_layers = 1')
        self.dropout_layer = Dropout(dropout) if dropout else None

    @staticmethod
    def init_stacked_analog_lstm(
            num_layers: int,
            layer: Type,
            first_layer_args: Any,
            other_layer_args: Any
    ) -> ModuleList:
        """Construct a list of LSTMLayers over which to iterate.

        Args:
            num_layers: number of serially connected LSTM layers
            layer: RNN layer type (e.g. AnalogLSTMLayer)
            first_layer_args: RNNCell type, input_size, hidden_size, rpu_config, etc.
            other_layer_args: RNNCell type, hidden_size, hidden_size, rpu_config, etc.

        Returns:
            torch.nn.ModuleList, which is similar to a regular Python list,
            but where torch.nn.Module methods can be applied
        """
        layers = [layer(*first_layer_args)] \
            + [layer(*other_layer_args) for _ in range(num_layers - 1)]
        return ModuleList(layers)

    def get_zero_state(self, batch_size: int) -> List[Tensor]:
        """Returns a zeroed state.

        Args:
            batch_size: batch size of the input

        Returns:
           List of zeroed state tensors for each layer
        """
        return [lay.get_zero_state(batch_size) for lay in self.layers]

    def forward(  # pylint: disable=arguments-differ
            self,
            input: Tensor,  # pylint: disable=redefined-builtin
            states: List
    ) -> Tuple[Tensor, List]:

        # List[RNNState]: One state per layer.
        output_states = jit.annotate(List, [])
        output = input

        for i, rnn_layer in enumerate(self.layers):
            state = states[i]
            output, out_state = rnn_layer(output, state)
            # Apply the dropout layer except the last layer.
            if i < self.num_layers - 1 and self.dropout_layer is not None:
                output = self.dropout_layer(output)
            output_states += [out_state]

        return output, output_states


class AnalogRNN(AnalogSequential):
    """Modular RNN that uses analog tiles.

    Args:
        cell: type of Analog RNN cell (AnalogLSTMCell/AnalogGRUCell/AnalogVanillaRNNCell)
        input_size: in_features to W_{ih} matrix of first layer
        hidden_size: in_features and out_features for W_{hh} matrices
        bias: whether to use a bias row on the analog tile or not
        rpu_config: resistive processing unit configuration.
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights
        xavier: whether standard PyTorch LSTM weight
            initialization (default) or Xavier initialization
        num_layers: number of serially connected RNN layers
        bidir: if True, becomes a bidirectional RNN
        dropout: dropout applied to output of all RNN layers except last
    """
    # pylint: disable=abstract-method, too-many-arguments

    def __init__(
            self,
            cell: Type,
            input_size: int,
            hidden_size: int,
            bias: bool = True,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            xavier: bool = False,
            num_layers: int = 1,
            bidir: bool = False,
            dropout: float = 0.0
            ):
        super().__init__()

        if bidir:
            layer = AnalogBidirRNNLayer
            num_dirs = 2
        else:
            layer = AnalogRNNLayer
            num_dirs = 1

        self.rnn = ModularRNN(
            num_layers, layer, dropout,
            first_layer_args=[cell, input_size, hidden_size, bias,
                              rpu_config, realistic_read_write],
            other_layer_args=[cell, num_dirs*hidden_size, hidden_size, bias,
                              rpu_config, realistic_read_write])
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.reset_parameters(xavier)

    def init_layers(
            self,
            weight_init_fn: Callable,
            bias_init_fn: Optional[Callable] = None
    ) -> None:
        """Init the analog layers with custom functions.

        Args:
            weight_init_fn: in-place tensor function applied to weight of
                ``AnalogLinear`` layers
            bias_init_fn: in-place tensor function applied to bias of
                ``AnalogLinear`` layers

        Note:
            If no bias init function is provided the weight init
            function is taken for the bias as well.
        """
        def init_analog_layer(layer: AnalogModuleBase) -> None:
            """Init the weights and bias of an analog linear layer."""
            weight_init_fn(layer.weight.data)
            if layer.use_bias:
                if bias_init_fn is None:
                    weight_init_fn(layer.bias.data)
                else:
                    bias_init_fn(layer.bias.data)

            layer.set_weights(layer.weight, layer.bias)

        self._apply_to_analog(init_analog_layer)  # pylint: disable=protected-access

    def reset_parameters(self, xavier: bool = False) -> None:
        """Weight and bias initialization.

        Args:
            xavier: whether standard PyTorch LSTM weight
               initialization (default) or Xavier initialization
        """
        if xavier:
            self.init_layers(init.xavier_uniform_, init.zeros_)
        else:
            stdv = 1. / math.sqrt(self.hidden_size)
            self.init_layers(lambda x: x.uniform_(-stdv, stdv))

    def get_zero_state(self, batch_size: int) -> List[Tensor]:
        """Returns a zeroed RNN state based on cell type and layer type

        Args:
            batch_size: batch size of the input

        Returns:
           List of zeroed state tensors for each layer

        """
        return self.rnn.get_zero_state(batch_size)

    def forward(
            self,
            input: Tensor,  # pylint: disable=redefined-builtin
            states: Optional[List] = None
    ) -> Tuple[Tensor, List]:
        if states is None:
            # TODO: batch_first.
            states = self.get_zero_state(input.shape[1])

        return self.rnn(input, states)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=too-many-locals, too-many-public-methods, no-member
"""Test for different utility functionality."""

from tempfile import TemporaryFile
from copy import deepcopy
from unittest import SkipTest

from numpy import array
from numpy.random import rand
from numpy.testing import assert_array_almost_equal, assert_raises
from torch import Tensor, load, save, device, manual_seed
from torch import abs as torch_abs
from torch.nn import Module, Sequential
from torch.nn import Linear as torch_linear
from torch.nn.functional import mse_loss
from torch.optim import SGD

from aihwkit.nn import (
    AnalogConv2d, AnalogConv2dMapped, AnalogSequential, AnalogLinearMapped
)
from aihwkit.optim import AnalogSGD
from aihwkit.simulator.configs import SingleRPUConfig, InferenceRPUConfig, FloatingPointRPUConfig
from aihwkit.simulator.configs.devices import ConstantStepDevice, LinearStepDevice
from aihwkit.simulator.configs.utils import IOParameters, UpdateParameters, MappingParameter
from aihwkit.exceptions import TileError, ModuleError
from aihwkit.nn.conversion import convert_to_analog

from .helpers.decorators import parametrize_over_layers
from .helpers.layers import (
    Conv2d, Conv2dCuda, Linear, LinearCuda, LinearMapped,
    LinearMappedCuda, Conv2dMapped, Conv2dMappedCuda
)
from .helpers.testcases import ParametrizedTestCase, SKIP_CUDA_TESTS
from .helpers.tiles import FloatingPoint, ConstantStep, Inference, InferenceLearnOutScaling


@parametrize_over_layers(
    layers=[Linear, Conv2d, LinearMapped, LinearCuda, LinearMappedCuda,
            Conv2dCuda, Conv2dMapped, Conv2dMappedCuda],
    tiles=[FloatingPoint, ConstantStep, Inference, InferenceLearnOutScaling],
    biases=['analog', 'digital', None],
)
class SerializationTest(ParametrizedTestCase):
    """Tests for serialization."""

    @staticmethod
    def train_model(model, loss_func, x_b, y_b):
        """Train the model."""
        opt = AnalogSGD(model.parameters(), lr=0.5)
        # opt.regroup_param_groups(model)
        epochs = 3
        for _ in range(epochs):
            opt.zero_grad()
            pred = model(x_b)

            loss = loss_func(pred, y_b)
            loss.backward()
            opt.step()
        return loss

    @staticmethod
    def get_layer_and_tile_weights(model):
        """Return the weights and biases of the model and the tile and whether
        it automatically syncs"""

        if isinstance(model, AnalogLinearMapped):
            weight, bias = model.get_weights()
            return weight, bias, weight, bias, True

        if isinstance(model, AnalogConv2dMapped):
            weight, bias = model.get_weights()
            return weight, bias, weight, bias, True

        weight = model.weight.data.detach().cpu().numpy()
        if model.use_bias:
            bias = model.bias.data.detach().cpu().numpy()
        else:
            bias = None

        analog_weight, analog_bias = model.get_weights()
        analog_weight = analog_weight.detach().cpu().numpy()
        if analog_bias is not None:
            analog_bias = analog_bias.detach().cpu().numpy()

        return weight, bias, analog_weight.reshape(weight.shape), analog_bias, True

    @staticmethod
    def get_analog_tile(model):
        """ Return a (python) analog tile of the model"""
        return list(model.analog_tiles())[0]

    def test_save_load_state_dict_train(self):
        """Test saving and loading using a state dict after training."""
        model = self.get_layer()

        # Perform an update in order to modify tile weights and biases.
        loss_func = mse_loss
        if isinstance(model, (AnalogConv2d, AnalogConv2dMapped)):
            input_x = Tensor(rand(2, 2, 3, 3))*0.2
            input_y = Tensor(rand(2, 3, 5, 5))*0.2
        else:
            input_x = Tensor(rand(2, model.in_features))*0.2
            input_y = Tensor(rand(2, model.out_features))*0.2

        if self.use_cuda:
            input_x = input_x.cuda()
            input_y = input_y.cuda()

        self.train_model(model, loss_func, input_x, input_y)

        # Keep track of the current weights and biases for comparing.
        (model_weights, model_biases,
         tile_weights, tile_biases, sync) = self.get_layer_and_tile_weights(model)

        # now the tile weights should be out of sync
        if not sync:
            assert_raises(AssertionError, assert_array_almost_equal, model_weights, tile_weights)
            if self.bias:
                assert_raises(AssertionError, assert_array_almost_equal, model_biases, tile_biases)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model.state_dict(), file)
            # Create a new model and load its state dict.
            file.seek(0)
            new_model = self.get_layer()
            new_model.load_state_dict(load(file))

        # Compare the new model weights and biases. they should now be in sync
        (new_model_weights, new_model_biases,
         new_tile_weights, new_tile_biases, _) = self.get_layer_and_tile_weights(new_model)

        assert_array_almost_equal(tile_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(tile_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

    def test_save_load_state_dict_train_after(self):
        """Test saving and loading using a state dict and training after load."""
        model = self.get_layer()

        # Perform an update in order to modify tile weights and biases.
        loss_func = mse_loss
        if isinstance(model, (AnalogConv2d, AnalogConv2dMapped)):
            input_x = Tensor(rand(2, 2, 3, 3))*0.2
            input_y = Tensor(rand(2, 3, 5, 5))*0.2
        else:
            input_x = Tensor(rand(2, model.in_features))*0.2
            input_y = Tensor(rand(2, model.out_features))*0.2

        if self.use_cuda:
            input_x = input_x.cuda()
            input_y = input_y.cuda()

        self.train_model(model, loss_func, input_x, input_y)

        # Keep track of the current weights and biases for comparing.
        _, _, tile_weights, tile_biases, _ = self.get_layer_and_tile_weights(model)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model.state_dict(), file)
            # Create a new model and load its state dict.
            file.seek(0)
            new_model = self.get_layer()
            new_model.load_state_dict(load(file))

        # Compare the new model weights and biases. they should now be in sync
        (new_model_weights, new_model_biases,
         new_tile_weights, new_tile_biases, _) = self.get_layer_and_tile_weights(new_model)

        assert_array_almost_equal(tile_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(tile_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

        new_loss = self.train_model(new_model, loss_func, input_x, input_y)

        loss = self.train_model(model, loss_func, input_x, input_y)
        if self.tile_class != ConstantStep:
            self.assertTensorAlmostEqual(loss, new_loss)

    def test_save_load_state_dict_train_after_old_model(self):
        """Test saving and loading using a state dict and training after load with old model."""
        model = self.get_layer()

        # Perform an update in order to modify tile weights and biases.
        loss_func = mse_loss
        if isinstance(model, (AnalogConv2d, AnalogConv2dMapped)):
            input_x = Tensor(rand(2, 2, 3, 3))*0.2
            input_y = Tensor(rand(2, 3, 5, 5))*0.2
        else:
            input_x = Tensor(rand(2, model.in_features))*0.2
            input_y = Tensor(rand(2, model.out_features))*0.2

        if self.use_cuda:
            input_x = input_x.cuda()
            input_y = input_y.cuda()

        self.train_model(model, loss_func, input_x, input_y)

        # Keep track of the current weights and biases for comparing.
        _, _, tile_weights, tile_biases, _ = self.get_layer_and_tile_weights(model)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model.state_dict(), file)
            # Create a new model and load its state dict.
            file.seek(0)
            model.load_state_dict(load(file))

        # Compare the new model weights and biases. they should now be in sync
        (new_model_weights, new_model_biases,
         new_tile_weights, new_tile_biases, _) = self.get_layer_and_tile_weights(model)

        assert_array_almost_equal(tile_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(tile_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

        self.train_model(model, loss_func, input_x, input_y)

    def test_save_load_train_after(self):
        """Test saving and loading using a state dict and training after load."""
        model = self.get_layer()

        # Perform an update in order to modify tile weights and biases.
        loss_func = mse_loss
        if isinstance(model, (AnalogConv2d, AnalogConv2dMapped)):
            input_x = Tensor(rand(2, 2, 3, 3))*0.2
            input_y = Tensor(rand(2, 3, 5, 5))*0.2
        else:
            input_x = Tensor(rand(2, model.in_features))*0.2
            input_y = Tensor(rand(2, model.out_features))*0.2

        if self.use_cuda:
            input_x = input_x.cuda()
            input_y = input_y.cuda()

        # Keep track of the current weights and biases for comparing.
        _, _, tile_weights, tile_biases, _ = self.get_layer_and_tile_weights(model)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Create a new model and load its state dict.
            file.seek(0)
            new_model = load(file)

        # Compare the new model weights and biases. they should now be in sync
        (new_model_weights, new_model_biases,
         new_tile_weights, new_tile_biases, _) = self.get_layer_and_tile_weights(new_model)

        assert_array_almost_equal(tile_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(tile_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

        new_loss = self.train_model(new_model, loss_func, input_x, input_y)
        loss = self.train_model(model, loss_func, input_x, input_y)
        if self.tile_class != ConstantStep:
            self.assertTensorAlmostEqual(loss, new_loss)

    def test_save_load_model(self):
        """Test saving and loading a model directly."""
        model = self.get_layer()

        # Keep track of the current weights and biases for comparing.
        (model_weights, model_biases,
         tile_weights, tile_biases, _) = self.get_layer_and_tile_weights(model)
        assert_array_almost_equal(model_weights, tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, tile_biases)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Compare the new model weights and biases.
        (new_model_weights, new_model_biases,
         new_tile_weights, new_tile_biases, _) = self.get_layer_and_tile_weights(new_model)

        assert_array_almost_equal(model_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

        # Asserts over the AnalogContext of the new model.
        new_analog_tile = self.get_analog_tile(new_model)
        analog_tile = self.get_analog_tile(model)
        self.assertTrue(hasattr(new_analog_tile.analog_ctx, 'analog_tile'))
        self.assertIsInstance(new_analog_tile.analog_ctx.analog_tile,
                              analog_tile.__class__)
        self.assertTrue(new_analog_tile.is_cuda == analog_tile.is_cuda)

    def test_save_load_model_cross_device(self):
        """Test saving and loading a model directly."""

        if SKIP_CUDA_TESTS:
            raise SkipTest('CUDA not available.')

        model = self.get_layer()

        map_location = 'cuda'
        if self.use_cuda:
            map_location = 'cpu'

        # Keep track of the current weights and biases for comparing.
        (model_weights, model_biases,
         tile_weights, tile_biases, _) = self.get_layer_and_tile_weights(model)
        assert_array_almost_equal(model_weights, tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, tile_biases)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file, map_location=device(map_location))

        # Compare the new model weights and biases.
        (new_model_weights, new_model_biases,
         new_tile_weights, new_tile_biases, _) = self.get_layer_and_tile_weights(new_model)

        assert_array_almost_equal(model_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

        # Asserts over the AnalogContext of the new model.
        new_analog_tile = self.get_analog_tile(new_model)
        analog_tile = self.get_analog_tile(model)

        self.assertTrue(hasattr(new_analog_tile.analog_ctx, 'analog_tile'))
        self.assertIsInstance(new_analog_tile.analog_ctx.analog_tile,
                              analog_tile.__class__)

        self.assertTrue(new_analog_tile.is_cuda != analog_tile.is_cuda)

        if analog_tile.shared_weights is not None:
            self.assertTrue(new_analog_tile.shared_weights.device.type == map_location)

    def test_save_load_meta_parameter(self):
        """Test saving and loading a device with custom parameters."""
        # Create the device and the array.
        rpu_config = SingleRPUConfig(
            forward=IOParameters(inp_noise=0.321),
            backward=IOParameters(inp_noise=0.456),
            update=UpdateParameters(desired_bl=78),
            device=ConstantStepDevice(w_max=0.987)
        )

        model = self.get_layer(rpu_config=rpu_config)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Assert over the new model tile parameters.
        new_analog_tile = self.get_analog_tile(new_model)
        analog_tile = self.get_analog_tile(model)

        parameters = new_analog_tile.tile.get_parameters()
        self.assertAlmostEqual(parameters.forward_io.inp_noise, 0.321)
        self.assertAlmostEqual(parameters.backward_io.inp_noise, 0.456)
        self.assertAlmostEqual(parameters.update.desired_bl, 78)
        self.assertTrue(new_analog_tile.is_cuda == analog_tile.is_cuda)

    def test_save_load_hidden_parameters(self):
        """Test saving and loading a device with hidden parameters."""
        # Create the device and the array.
        model = self.get_layer()
        analog_tile = self.get_analog_tile(model)
        hidden_parameters = analog_tile.tile.get_hidden_parameters()

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Assert over the new model tile parameters.
        new_analog_tile = self.get_analog_tile(new_model)
        new_hidden_parameters = new_analog_tile.tile.get_hidden_parameters()
        assert_array_almost_equal(hidden_parameters, new_hidden_parameters)

    def test_save_load_out_scaling_alpha(self):
        """Test saving and loading a device with out_scaling_alpha."""
        # Create the device and the array.
        model = self.get_layer()
        alpha = 2.0
        analog_tile = self.get_analog_tile(model)
        analog_tile.out_scaling_alpha = Tensor([alpha])

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Assert over the new model tile parameters.
        new_analog_tile = self.get_analog_tile(new_model)
        alpha_new = new_analog_tile.get_scales().detach().cpu()
        assert_array_almost_equal(array(alpha), array(alpha_new))

    def test_save_load_shared_weights(self):
        """Test saving and loading a device with shared_weights."""

        if isinstance(self.get_rpu_config(), FloatingPointRPUConfig):
            raise SkipTest('Not available for FP')

        # Create the device and the array.
        model = self.get_layer()
        analog_tile = self.get_analog_tile(model)

        shared_weights = None
        if analog_tile.shared_weights is not None:
            shared_weights = analog_tile.shared_weights.detach().cpu().numpy()

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Assert over the new model tile parameters.
        new_analog_tile = self.get_analog_tile(new_model)
        if shared_weights is not None:
            new_shared_weights = new_analog_tile.shared_weights
            assert_array_almost_equal(shared_weights, new_shared_weights.detach().cpu().numpy())

    def test_save_load_weight_scaling_omega(self):
        """Test saving and loading a device with weight scaling omega."""
        # Create the device and the array.
        rpu_config = SingleRPUConfig(mapping=MappingParameter(weight_scaling_omega=0.4))
        model = self.get_layer(rpu_config=rpu_config)
        analog_tile = self.get_analog_tile(model)
        alpha = analog_tile.get_scales().detach().cpu()
        self.assertNotEqual(alpha, 1.0)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Assert over the new model tile parameters.
        new_analog_tile = self.get_analog_tile(new_model)
        alpha_new = new_analog_tile.get_scales().detach().cpu()
        assert_array_almost_equal(array(alpha), array(alpha_new))

    def test_remapping(self):
        """Test remapping of the weights."""
        # Create the device and the array.
        rpu_config = InferenceRPUConfig(mapping=MappingParameter(
            weight_scaling_omega=0.4,
            weight_scaling_columnwise=True))

        model = self.get_layer(rpu_config=rpu_config)
        analog_tile = self.get_analog_tile(model)

        user_weights = Tensor(model.out_features, model.in_features).uniform_(-0.1, 0.1)
        if self.bias:
            user_biases = Tensor(model.out_features).uniform_(-0.1, 0.1)
        else:
            user_biases = None
        model.set_weights(user_weights, user_biases, apply_weight_scaling=True,
                          force_exact=True)

        alpha_initial = analog_tile.get_scales().detach().cpu()
        self.assertNotEqual(alpha_initial.max(), 1.0)

        # remap module
        model.remap_weights(weight_scaling_omega=1.0)

        weights, _ = model.get_weights(apply_weight_scaling=True, force_exact=True)
        analog_weights, _ = model.get_weights(apply_weight_scaling=False, force_exact=True)

        assert_array_almost_equal(array(user_weights), array(weights))

        alpha = analog_tile.get_scales().detach().cpu()
        assert_raises(AssertionError, assert_array_almost_equal, array(alpha), array(alpha_initial))

        self.assertEqual(torch_abs(analog_weights).max(), 1.0)

    def test_save_load_state_dict_hidden_parameters(self):
        """Test saving and loading via state_dict with hidden parameters."""
        # Create the device and the array.
        model = self.get_layer()
        analog_tile = self.get_analog_tile(model)
        hidden_parameters = analog_tile.tile.get_hidden_parameters()

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model.state_dict(), file)
            # Load the model.
            file.seek(0)
            new_model = self.get_layer()
            new_model.load_state_dict(load(file))

        # Assert over the new model tile parameters.
        new_analog_tile = self.get_analog_tile(new_model)
        new_hidden_parameters = new_analog_tile.tile.get_hidden_parameters()
        assert_array_almost_equal(hidden_parameters, new_hidden_parameters)

    def test_state_dict_children_layers_sequential(self):
        """Test using the state_dict with children analog layers via Sequential."""
        children_layer = self.get_layer()
        model = Sequential(children_layer)

        # Keep track of the current weights and biases for comparing.
        (model_weights, model_biases,
         tile_weights, tile_biases, _) = self.get_layer_and_tile_weights(children_layer)

        state_dict = model.state_dict()
        lst = [key for key in state_dict.keys() if key.startswith('0.')
               and children_layer.ANALOG_STATE_PREFIX in key]
        self.assertTrue(len(lst) > 0)

        # Update the state_dict of a new model.
        new_children_layer = self.get_layer()
        new_model = Sequential(new_children_layer)
        new_model.load_state_dict(model.state_dict())

        # Compare the new model weights and biases.
        (new_model_weights, new_model_biases, new_tile_weights, new_tile_biases, _) = \
            self.get_layer_and_tile_weights(new_children_layer)

        assert_array_almost_equal(model_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

    def test_state_dict_children_layers_subclassing(self):
        """Test using the state_dict with children analog layers via subclassing."""

        class CustomModule(Module):
            """Module that defines its children layers via custom attributes."""
            # pylint: disable=abstract-method
            def __init__(self, layer: Module):
                super().__init__()
                self.custom_child = layer

        children_layer = self.get_layer()
        model = CustomModule(children_layer)

        # Keep track of the current weights and biases for comparing.
        (model_weights, model_biases, tile_weights, tile_biases, _) = \
            self.get_layer_and_tile_weights(children_layer)

        state_dict = model.state_dict()
        lst = [key for key in state_dict.keys() if key.startswith('custom_child.')
               and children_layer.ANALOG_STATE_PREFIX in key]
        self.assertTrue(len(lst) > 0)

        # Update the state_dict of a new model.
        new_children_layer = self.get_layer()
        new_model = CustomModule(new_children_layer)
        new_model.load_state_dict(model.state_dict())

        # Compare the new model weights and biases.
        (new_model_weights, new_model_biases, new_tile_weights, new_tile_biases, _) = \
            self.get_layer_and_tile_weights(new_children_layer)

        assert_array_almost_equal(model_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

    def test_state_dict_analog_strict(self):
        """Test the `strict` flag for analog layers."""
        model = self.get_layer()
        state_dict = model.state_dict()

        # Remove the analog key from the state dict.
        lst = [key for key in state_dict.keys() if key.startswith(model.ANALOG_STATE_PREFIX)]
        del state_dict[lst[0]]

        # Check that it fails when using `strict`.
        with self.assertRaises(RuntimeError) as context:
            model.load_state_dict(state_dict, strict=True)
        self.assertIn('Missing key', str(context.exception))

        # Check that it passes when not using `strict`.
        model.load_state_dict(state_dict, strict=False)

    def test_state_dict(self):
        """Test creating a new model using a state dict, without saving to disk."""
        model = self.get_layer()
        state_dict = model.state_dict()
        analog_tile = self.get_analog_tile(model)

        new_model = self.get_layer()
        new_model.load_state_dict(state_dict)
        new_analog_tile = self.get_analog_tile(new_model)

        # Asserts over the AnalogContext of the new model.
        self.assertTrue(hasattr(new_analog_tile.analog_ctx, 'analog_tile'))
        self.assertIsInstance(new_analog_tile.analog_ctx.analog_tile,
                              analog_tile.__class__)

    def test_hidden_parameter_mismatch(self):
        """Test for error if tile structure mismatches."""
        model = self.get_layer()
        state_dict = model.state_dict()
        analog_tile = self.get_analog_tile(model)

        # Create the device and the array.
        rpu_config = SingleRPUConfig(
            device=LinearStepDevice()  # different hidden structure
        )

        new_model = self.get_layer(rpu_config=rpu_config)
        new_analog_tile = self.get_analog_tile(new_model)
        if new_analog_tile.__class__.__name__ != analog_tile.__class__.__name__:
            with self.assertRaises(TileError):
                new_model.load_state_dict(state_dict)

    def test_load_state_load_rpu_config(self):
        """Test creating a new model using a state dict, while using a different RPU config."""

        # Create the device and the array.
        rpu_config_org = self.get_rpu_config()

        # Skipped for FP
        if isinstance(rpu_config_org, FloatingPointRPUConfig):
            raise SkipTest('Not available for FP')

        rpu_config_org.forward.is_perfect = False
        old_value = 0.11
        rpu_config_org.forward.inp_noise = old_value

        model = self.get_layer(rpu_config=rpu_config_org)
        state_dict = model.state_dict()

        rpu_config = deepcopy(rpu_config_org)
        new_value = 0.51
        rpu_config.forward.inp_noise = new_value

        # Test restore_rpu_config=False
        new_model = self.get_layer(rpu_config=rpu_config)
        new_model.load_state_dict(state_dict, load_rpu_config=False)
        new_analog_tile = self.get_analog_tile(new_model)

        parameters = new_analog_tile.tile.get_parameters()
        self.assertAlmostEqual(parameters.forward_io.inp_noise, new_value)

        # Test restore_rpu_config=True
        new_model = self.get_layer(rpu_config=rpu_config)
        new_model.load_state_dict(state_dict, load_rpu_config=True)
        new_analog_tile = self.get_analog_tile(new_model)

        parameters = new_analog_tile.tile.get_parameters()
        self.assertAlmostEqual(parameters.forward_io.inp_noise, old_value)

    def test_load_state_load_rpu_config_sequential(self):
        """Test creating a new model using a state dict, while using a different RPU config."""

        # Create the device and the array.
        rpu_config_org = self.get_rpu_config()

        # Skipped for FP
        if isinstance(rpu_config_org, FloatingPointRPUConfig):
            raise SkipTest('Not available for FP')

        rpu_config_org.forward.is_perfect = False
        old_value = 0.11
        rpu_config_org.forward.inp_noise = old_value

        model = AnalogSequential(self.get_layer(rpu_config=rpu_config_org))
        state_dict = model.state_dict()

        rpu_config = deepcopy(rpu_config_org)
        new_value = 0.51
        rpu_config.forward.inp_noise = new_value

        # Test restore_rpu_config=False
        new_model = AnalogSequential(self.get_layer(rpu_config=rpu_config))
        new_model.load_state_dict(state_dict, load_rpu_config=False)
        new_analog_tile = self.get_analog_tile(new_model[0])

        parameters = new_analog_tile.tile.get_parameters()
        self.assertAlmostEqual(parameters.forward_io.inp_noise, new_value)

        # Test restore_rpu_config=True
        new_model = AnalogSequential(self.get_layer(rpu_config=rpu_config))
        new_model.load_state_dict(state_dict, load_rpu_config=True)
        new_analog_tile = self.get_analog_tile(new_model[0])

        parameters = new_analog_tile.tile.get_parameters()
        self.assertAlmostEqual(parameters.forward_io.inp_noise, old_value)

    def test_load_state_load_rpu_config_wrong(self):
        """Test creating a new model using a state dict, while using a different RPU config."""

        # Skipped for FP
        if isinstance(self.get_rpu_config(), FloatingPointRPUConfig):
            raise SkipTest('Not available for FP')

        # Create the device and the array.
        model = self.get_layer()
        state_dict = model.state_dict()

        rpu_config = FloatingPointRPUConfig()

        new_model = self.get_layer(rpu_config=rpu_config)
        assert_raises(ModuleError, new_model.load_state_dict, state_dict, load_rpu_config=False)


@parametrize_over_layers(
    layers=[Linear, LinearCuda],
    tiles=[FloatingPoint],
    biases=[None],
)
class SerializationTestExtended(ParametrizedTestCase):
    """Tests for serialization."""

    @staticmethod
    def train_model_torch(model, loss_func, x_b, y_b):
        """Train the model with torch SGD."""
        opt = SGD(model.parameters(), lr=0.5)
        epochs = 100
        for _ in range(epochs):
            opt.zero_grad()
            pred = model(x_b)
            loss = loss_func(pred, y_b)

            loss.backward()
            opt.step()

    @staticmethod
    def get_torch_model(use_cuda: bool):
        """ Returns a torch model."""
        manual_seed(4321)
        torch_model = Sequential(
            torch_linear(4, 3),
            torch_linear(3, 3),
            Sequential(
                torch_linear(3, 1),
                torch_linear(1, 1)
            )
        )
        if use_cuda:
            torch_model.cuda()
        return torch_model

    def test_load_state_dict_conversion(self):
        """Test loading and setting conversion with alpha."""

        # Create the device and the array.

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])
        y_b = Tensor([[0.3], [0.6]])
        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()

        model = self.get_torch_model(self.use_cuda)

        self.train_model_torch(model, mse_loss, x_b, y_b)

        analog_model = convert_to_analog(model, self.get_rpu_config())
        analog_loss = mse_loss(analog_model(x_b), y_b)

        with TemporaryFile() as file:
            save(analog_model.state_dict(), file)
            # Load the model.
            file.seek(0)
            model = self.get_torch_model(self.use_cuda)
            new_analog_model = convert_to_analog(model, self.get_rpu_config())
            state_dict = load(file)
            new_analog_model.load_state_dict(state_dict, load_rpu_config=True)

        new_state_dict = new_analog_model.state_dict()
        for key in new_state_dict.keys():

            if not key.endswith('analog_tile_state'):
                continue

            state1 = new_state_dict[key]
            state2 = state_dict[key]
            assert_array_almost_equal(state1['analog_tile_weights'],
                                      state2['analog_tile_weights'])
            assert_array_almost_equal(state1['analog_alpha_scale'],
                                      state2['analog_alpha_scale'])

        new_analog_loss = mse_loss(new_analog_model(x_b), y_b)
        self.assertTensorAlmostEqual(new_analog_loss, analog_loss)
